"""
Tristimulus Values
==================

Defines the objects for tristimulus values computation from spectral data:

-   :attr:`colour.SPECTRAL_SHAPE_ASTME308`
-   :func:`colour.colorimetry.handle_spectral_arguments`
-   :func:`colour.colorimetry.tristimulus_weighting_factors_ASTME2022`
-   :func:`colour.colorimetry.sd_to_XYZ_integration`
-   :func:`colour.colorimetry.\
sd_to_XYZ_tristimulus_weighting_factors_ASTME308`
-   :func:`colour.colorimetry.sd_to_XYZ_ASTME308`
-   :attr:`colour.SD_TO_XYZ_METHODS`
-   :func:`colour.sd_to_XYZ`
-   :func:`colour.colorimetry.msds_to_XYZ_integration`
-   :func:`colour.colorimetry.msds_to_XYZ_ASTME308`
-   :attr:`colour.MSDS_TO_XYZ_METHODS`
-   :func:`colour.msds_to_XYZ`
-   :func:`colour.wavelength_to_XYZ`

The default implementation is based on practise *ASTM E308-15* method.

References
----------
-   :cite:`ASTMInternational2011a` : ASTM International. (2011). ASTM E2022-11
    - Standard Practice for Calculation of Weighting Factors for Tristimulus
    Integration (pp. 1-10). doi:10.1520/E2022-11
-   :cite:`ASTMInternational2015b` : ASTM International. (2015). ASTM E308-15 -
    Standard Practice for Computing the Colors of Objects by Using the CIE
    System (pp. 1-47). doi:10.1520/E0308-15
-   :cite:`Wyszecki2000bf` : Wyszecki, Günther, & Stiles, W. S. (2000).
    Integration Replaced by Summation. In Color Science: Concepts and Methods,
    Quantitative Data and Formulae (pp. 158-163). Wiley. ISBN:978-0-471-39918-6
"""

from __future__ import annotations

import numpy as np

from colour.algebra import lagrange_coefficients
from colour.colorimetry import (
    SPECTRAL_SHAPE_DEFAULT,
    MultiSpectralDistributions,
    SpectralDistribution,
    SpectralShape,
    reshape_msds,
    reshape_sd,
)
from colour.hints import (
    Any,
    ArrayLike,
    Boolean,
    Dict,
    FloatingOrNDArray,
    Integer,
    Literal,
    NDArray,
    Number,
    Optional,
    Tuple,
    Union,
    cast,
)
from colour.utilities import (
    CACHE_REGISTRY,
    CaseInsensitiveMapping,
    as_float_array,
    as_int_scalar,
    attest,
    filter_kwargs,
    from_range_100,
    get_domain_range_scale,
    optional,
    runtime_warning,
    validate_method,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "SPECTRAL_SHAPE_ASTME308",
    "handle_spectral_arguments",
    "lagrange_coefficients_ASTME2022",
    "tristimulus_weighting_factors_ASTME2022",
    "adjust_tristimulus_weighting_factors_ASTME308",
    "sd_to_XYZ_integration",
    "sd_to_XYZ_tristimulus_weighting_factors_ASTME308",
    "sd_to_XYZ_ASTME308",
    "SD_TO_XYZ_METHODS",
    "sd_to_XYZ",
    "msds_to_XYZ_integration",
    "msds_to_XYZ_ASTME308",
    "MSDS_TO_XYZ_METHODS",
    "msds_to_XYZ",
    "wavelength_to_XYZ",
]

SPECTRAL_SHAPE_ASTME308: SpectralShape = SPECTRAL_SHAPE_DEFAULT
SPECTRAL_SHAPE_ASTME308.__doc__ = """
Shape for *ASTM E308-15* practise: (360, 780, 1).

References
----------
:cite:`ASTMInternational2015b`
"""

_CACHE_LAGRANGE_INTERPOLATING_COEFFICIENTS: Dict = (
    CACHE_REGISTRY.register_cache(
        f"{__name__}._CACHE_LAGRANGE_INTERPOLATING_COEFFICIENTS"
    )
)

_CACHE_TRISTIMULUS_WEIGHTING_FACTORS: Dict = CACHE_REGISTRY.register_cache(
    f"{__name__}._CACHE_TRISTIMULUS_WEIGHTING_FACTORS"
)

_CACHE_SD_TO_XYZ: Dict = CACHE_REGISTRY.register_cache(
    f"{__name__}._CACHE_SD_TO_XYZ"
)


def handle_spectral_arguments(
    cmfs: Optional[MultiSpectralDistributions] = None,
    illuminant: Optional[SpectralDistribution] = None,
    cmfs_default: str = "CIE 1931 2 Degree Standard Observer",
    illuminant_default: str = "D65",
    shape_default: SpectralShape = SPECTRAL_SHAPE_DEFAULT,
    issue_runtime_warnings: bool = True,
) -> Tuple[MultiSpectralDistributions, SpectralDistribution]:
    """
    Handle the spectral arguments of various *Colour* definitions performing
    spectral computations.

    -   If ``cmfs`` is not given, one is chosen according to ``cmfs_default``.
        The returned colour matching functions adopt the spectral shape given
        by ``shape_default``.
    -   If ``illuminant`` is not given, one is chosen according to
        ``illuminant_default``. The returned illuminant adopts the spectral
        shape of the returned colour matching functions.
    -   If ``illuminant`` is given, the returned illuminant spectral shape is
        aligned to that of the returned colour matching functions.

    Parameters
    ----------
    cmfs
        Standard observer colour matching functions, default to the
        *CIE 1931 2 Degree Standard Observer*.
    illuminant
        Illuminant spectral distribution, default to
        *CIE Standard Illuminant D65*.
    cmfs_default
        The default colour matching functions to use if ``cmfs`` is not given.
    illuminant_default
        The default illuminant to use if ``illuminant`` is not given.
    shape_default
        The default spectral shape to align the final colour matching functions
        and illuminant.
    issue_runtime_warnings
        Whether to issue the runtime warnings.

    Returns
    -------
    :class:`tuple`
        Colour matching functions and illuminant.

    Examples
    --------
    >>> cmfs, illuminant = handle_spectral_arguments()
    >>> cmfs.name, cmfs.shape, illuminant.name, illuminant.shape
    ('CIE 1931 2 Degree Standard Observer', SpectralShape(360.0, 780.0, 1.0), \
'D65', SpectralShape(360.0, 780.0, 1.0))
    >>> cmfs, illuminant = handle_spectral_arguments(
    ...     shape_default=SpectralShape(400, 700, 20))
    >>> cmfs.name, cmfs.shape, illuminant.name, illuminant.shape
    ('CIE 1931 2 Degree Standard Observer', \
SpectralShape(400.0, 700.0, 20.0), 'D65', SpectralShape(400.0, 700.0, 20.0))
    """

    from colour import MSDS_CMFS, SDS_ILLUMINANTS

    cmfs = optional(cmfs, reshape_msds(MSDS_CMFS[cmfs_default], shape_default))
    illuminant = optional(
        illuminant, reshape_sd(SDS_ILLUMINANTS[illuminant_default], cmfs.shape)
    )

    if illuminant.shape != cmfs.shape:
        issue_runtime_warnings and runtime_warning(
            f'Aligning "{illuminant.name}" illuminant shape to "{cmfs.name}" '
            f"colour matching functions shape."
        )

        illuminant = reshape_sd(illuminant, cmfs.shape)

    return cmfs, illuminant


def lagrange_coefficients_ASTME2022(
    interval: Integer = 10,
    interval_type: Union[Literal["Boundary", "Inner"], str] = "Inner",
) -> NDArray:
    """
    Compute the *Lagrange Coefficients* for given interval size using practise
    *ASTM E2022-11* method.

    Parameters
    ----------
    interval
        Interval size in nm.
    interval_type
        If the interval is an *inner* interval *Lagrange Coefficients* are
        computed for degree 4. Degree 3 is used for a *boundary* interval.

    Returns
    -------
    :class:`numpy.ndarray`
        *Lagrange Coefficients*.

    References
    ----------
    :cite:`ASTMInternational2011a`

    Examples
    --------
    >>> lagrange_coefficients_ASTME2022(10, 'inner')
    ... # doctest: +ELLIPSIS
    array([[-0.028...,  0.940...,  0.104..., -0.016...],
           [-0.048...,  0.864...,  0.216..., -0.032...],
           [-0.059...,  0.773...,  0.331..., -0.045...],
           [-0.064...,  0.672...,  0.448..., -0.056...],
           [-0.062...,  0.562...,  0.562..., -0.062...],
           [-0.056...,  0.448...,  0.672..., -0.064...],
           [-0.045...,  0.331...,  0.773..., -0.059...],
           [-0.032...,  0.216...,  0.864..., -0.048...],
           [-0.016...,  0.104...,  0.940..., -0.028...]])
    >>> lagrange_coefficients_ASTME2022(10, 'boundary')
    ... # doctest: +ELLIPSIS
    array([[ 0.85...,  0.19..., -0.04...],
           [ 0.72...,  0.36..., -0.08...],
           [ 0.59...,  0.51..., -0.10...],
           [ 0.48...,  0.64..., -0.12...],
           [ 0.37...,  0.75..., -0.12...],
           [ 0.28...,  0.84..., -0.12...],
           [ 0.19...,  0.91..., -0.10...],
           [ 0.12...,  0.96..., -0.08...],
           [ 0.05...,  0.99..., -0.04...]])
    """

    global _CACHE_LAGRANGE_INTERPOLATING_COEFFICIENTS

    interval_type = validate_method(
        interval_type,
        ["Boundary", "Inner"],
        '"{0}" interval type is invalid, it must be one of {1}!',
    )

    hash_key = tuple(hash(arg) for arg in (interval, interval_type))
    if hash_key in _CACHE_LAGRANGE_INTERPOLATING_COEFFICIENTS:
        return np.copy(_CACHE_LAGRANGE_INTERPOLATING_COEFFICIENTS[hash_key])

    r_n = np.linspace(1 / interval, 1 - (1 / interval), interval - 1)
    d = 3
    if interval_type == "inner":
        r_n += 1
        d = 4

    lica = as_float_array([lagrange_coefficients(r, d) for r in r_n])

    _CACHE_LAGRANGE_INTERPOLATING_COEFFICIENTS[hash_key] = np.copy(lica)

    return lica


def tristimulus_weighting_factors_ASTME2022(
    cmfs: MultiSpectralDistributions,
    illuminant: SpectralDistribution,
    shape: SpectralShape,
    k: Optional[Number] = None,
) -> NDArray:
    """
    Return a table of tristimulus weighting factors for given colour matching
    functions and illuminant using practise *ASTM E2022-11* method.

    The computed table of tristimulus weighting factors should be used with
    spectral data that has been corrected for spectral bandpass dependence.

    Parameters
    ----------
    cmfs
        Standard observer colour matching functions.
    illuminant
        Illuminant spectral distribution.
    shape
        Shape used to build the table, only the interval is needed.
    k
        Normalisation constant :math:`k`. For reflecting or transmitting object
        colours, :math:`k` is chosen so that :math:`Y = 100` for objects for
        which the spectral reflectance factor :math:`R(\\lambda)` of the object
        colour or the spectral transmittance factor :math:`\\tau(\\lambda)` of
        the object is equal to unity for all wavelengths. For self-luminous
        objects and illuminants, the constants :math:`k` is usually chosen on
        the grounds of convenience. If, however, in the CIE 1931 standard
        colorimetric system, the :math:`Y` value is required to be numerically
        equal to the absolute value of a photometric quantity, the constant,
        :math:`k`, must be put equal to the numerical value of :math:`K_m`, the
        maximum spectral luminous efficacy (which is equal to
        683 :math:`lm\\cdot W^{-1}`) and :math:`\\Phi_\\lambda(\\lambda)` must
        be the spectral concentration of the radiometric quantity corresponding
        to the photometric quantity required.

    Returns
    -------
    :class:`numpy.ndarray`
        Tristimulus weighting factors table.

    Raises
    ------
    ValueError
        If the colour matching functions or illuminant intervals are not equal
        to 1 nm.

    Notes
    -----
    -   Input colour matching functions and illuminant intervals are expected
        to be equal to 1 nm. If the illuminant data is not available at 1 nm
        interval, it needs to be interpolated using *CIE* recommendations:
        The method developed by *Sprague (1880)* should be used for
        interpolating functions having a uniformly spaced independent variable
        and a *Cubic Spline* method for non-uniformly spaced independent
        variable.

    References
    ----------
    :cite:`ASTMInternational2011a`

    Examples
    --------
    >>> from colour import (MSDS_CMFS, SpectralDistribution, SpectralShape,
    ...     sd_CIE_standard_illuminant_A)
    >>> from colour.utilities import numpy_print_options
    >>> cmfs = MSDS_CMFS['CIE 1964 10 Degree Standard Observer']
    >>> A = sd_CIE_standard_illuminant_A(cmfs.shape)
    >>> with numpy_print_options(suppress=True):
    ...     tristimulus_weighting_factors_ASTME2022(
    ...         cmfs, A, SpectralShape(360, 830, 20))
    ... # doctest: +ELLIPSIS
    array([[ -0.0002981...,  -0.0000317...,  -0.0013301...],
           [ -0.0087155...,  -0.0008915...,  -0.0407436...],
           [  0.0599679...,   0.0050203...,   0.2565018...],
           [  0.7734225...,   0.0779839...,   3.6965732...],
           [  1.9000905...,   0.3037005...,   9.7554195...],
           [  1.9707727...,   0.8552809...,  11.4867325...],
           [  0.7183623...,   2.1457000...,   6.7845806...],
           [  0.0426667...,   4.8985328...,   2.3208000...],
           [  1.5223302...,   9.6471138...,   0.7430671...],
           [  5.6770329...,  14.4609708...,   0.1958194...],
           [ 12.4451744...,  17.4742541...,   0.0051827...],
           [ 20.5535772...,  17.5838219...,  -0.0026512...],
           [ 25.3315384...,  14.8957035...,   0.       ...],
           [ 21.5711570...,  10.0796619...,   0.       ...],
           [ 12.1785817...,   5.0680655...,   0.       ...],
           [  4.6675746...,   1.8303239...,   0.       ...],
           [  1.3236117...,   0.5129694...,   0.       ...],
           [  0.3175325...,   0.1230084...,   0.       ...],
           [  0.0746341...,   0.0290243...,   0.       ...],
           [  0.0182990...,   0.0071606...,   0.       ...],
           [  0.0047942...,   0.0018888...,   0.       ...],
           [  0.0013293...,   0.0005277...,   0.       ...],
           [  0.0004254...,   0.0001704...,   0.       ...],
           [  0.0000962...,   0.0000389...,   0.       ...]])
    """

    if cmfs.shape.interval != 1:
        raise ValueError(f'"{cmfs}" shape "interval" must be 1!')

    if illuminant.shape.interval != 1:
        raise ValueError(f'"{illuminant}" shape "interval" must be 1!')

    global _CACHE_TRISTIMULUS_WEIGHTING_FACTORS

    hash_key = tuple(
        hash(arg)
        for arg in (cmfs, illuminant, shape, k, get_domain_range_scale())
    )
    if hash_key in _CACHE_TRISTIMULUS_WEIGHTING_FACTORS:
        return np.copy(_CACHE_TRISTIMULUS_WEIGHTING_FACTORS[hash_key])

    Y = cmfs.values
    S = illuminant.values

    interval_i = int(shape.interval)
    W = S[::interval_i, np.newaxis] * Y[::interval_i, :]

    # First and last measurement intervals *Lagrange Coefficients*.
    c_c = lagrange_coefficients_ASTME2022(interval_i, "boundary")
    # Intermediate measurement intervals *Lagrange Coefficients*.
    c_b = lagrange_coefficients_ASTME2022(interval_i, "inner")

    # Total wavelengths count.
    w_c = len(Y)
    # Measurement interval interpolated values count.
    r_c = c_b.shape[0]
    # Last interval first interpolated wavelength.
    w_lif = w_c - (w_c - 1) % interval_i - 1 - r_c

    # Intervals count.
    i_c = W.shape[0]
    i_cm = i_c - 1

    # "k" is used as index in the nested loop.
    k_n = k

    for i in range(3):
        # First interval.
        for j in range(r_c):
            for k in range(3):
                W[k, i] = W[k, i] + c_c[j, k] * S[j + 1] * Y[j + 1, i]

        # Last interval.
        for j in range(r_c):
            for k in range(i_cm, i_cm - 3, -1):
                W[k, i] = (
                    W[k, i]
                    + c_c[r_c - j - 1, i_cm - k]
                    * S[j + w_lif]
                    * Y[j + w_lif, i]
                )

        # Intermediate intervals.
        for j in range(i_c - 3):
            for k in range(r_c):
                w_i = (r_c + 1) * (j + 1) + 1 + k
                W[j, i] = W[j, i] + c_b[k, 0] * S[w_i] * Y[w_i, i]
                W[j + 1, i] = W[j + 1, i] + c_b[k, 1] * S[w_i] * Y[w_i, i]
                W[j + 2, i] = W[j + 2, i] + c_b[k, 2] * S[w_i] * Y[w_i, i]
                W[j + 3, i] = W[j + 3, i] + c_b[k, 3] * S[w_i] * Y[w_i, i]

        # Extrapolation of potential incomplete interval.
        for j in range(as_int_scalar(w_c - ((w_c - 1) % interval_i)), w_c, 1):
            W[i_cm, i] = W[i_cm, i] + S[j] * Y[j, i]

    W *= cast(Number, optional(k_n, 100 / np.sum(W, axis=0)[1]))

    _CACHE_TRISTIMULUS_WEIGHTING_FACTORS[hash_key] = np.copy(W)

    return W


def adjust_tristimulus_weighting_factors_ASTME308(
    W: ArrayLike, shape_r: SpectralShape, shape_t: SpectralShape
) -> NDArray:
    """
    Adjust given table of tristimulus weighting factors to account for a
    shorter wavelengths range of the test spectral shape compared to the
    reference spectral shape using practise  *ASTM E308-15* method:
    Weights at the wavelengths for which data are not available are added to
    the weights at the shortest and longest wavelength for which spectral data
    are available.

    Parameters
    ----------
    W
        Tristimulus weighting factors table.
    shape_r
        Reference spectral shape.
    shape_t
        Test spectral shape.

    Returns
    -------
    :class:`numpy.ndarray`
        Adjusted tristimulus weighting factors.

    References
    ----------
    :cite:`ASTMInternational2015b`

    Examples
    --------
    >>> from colour import (MSDS_CMFS, SpectralDistribution, SpectralShape,
    ...     sd_CIE_standard_illuminant_A)
    >>> from colour.utilities import numpy_print_options
    >>> cmfs = MSDS_CMFS['CIE 1964 10 Degree Standard Observer']
    >>> A = sd_CIE_standard_illuminant_A(cmfs.shape)
    >>> W = tristimulus_weighting_factors_ASTME2022(
    ...     cmfs, A, SpectralShape(360, 830, 20))
    >>> with numpy_print_options(suppress=True):
    ...     adjust_tristimulus_weighting_factors_ASTME308(
    ...         W,  SpectralShape(360, 830, 20), SpectralShape(400, 700, 20))
    ... # doctest: +ELLIPSIS
    array([[  0.0509543...,   0.0040971...,   0.2144280...],
           [  0.7734225...,   0.0779839...,   3.6965732...],
           [  1.9000905...,   0.3037005...,   9.7554195...],
           [  1.9707727...,   0.8552809...,  11.4867325...],
           [  0.7183623...,   2.1457000...,   6.7845806...],
           [  0.0426667...,   4.8985328...,   2.3208000...],
           [  1.5223302...,   9.6471138...,   0.7430671...],
           [  5.6770329...,  14.4609708...,   0.1958194...],
           [ 12.4451744...,  17.4742541...,   0.0051827...],
           [ 20.5535772...,  17.5838219...,  -0.0026512...],
           [ 25.3315384...,  14.8957035...,   0.       ...],
           [ 21.5711570...,  10.0796619...,   0.       ...],
           [ 12.1785817...,   5.0680655...,   0.       ...],
           [  4.6675746...,   1.8303239...,   0.       ...],
           [  1.3236117...,   0.5129694...,   0.       ...],
           [  0.4171109...,   0.1618194...,   0.       ...]])
    """

    W = as_float_array(W).copy()

    start_index = int((shape_t.start - shape_r.start) / shape_r.interval)
    for i in range(start_index):
        W[start_index] += W[i]

    end_index = int((shape_r.end - shape_t.end) / shape_r.interval)
    for i in range(end_index):
        W[-end_index - 1] += W[-i - 1]

    return W[start_index : -end_index or None, ...]


def sd_to_XYZ_integration(
    sd: Union[ArrayLike, SpectralDistribution, MultiSpectralDistributions],
    cmfs: Optional[MultiSpectralDistributions] = None,
    illuminant: Optional[SpectralDistribution] = None,
    k: Optional[Number] = None,
    shape: Optional[SpectralShape] = None,
) -> NDArray:
    """
    Convert given spectral distribution to *CIE XYZ* tristimulus values
    using given colour matching functions and illuminant according to classical
    integration method.

    The spectral distribution can be either a
    :class:`colour.SpectralDistribution` class instance or an `ArrayLike` in
    which case the ``shape`` must be passed.

    Parameters
    ----------
    sd
        Spectral distribution, if an `ArrayLike` the wavelengths are
        expected to be in the last axis, e.g. for a spectral array with
        77 bins, ``sd`` shape could be (77, ) or (1, 77).
    cmfs
        Standard observer colour matching functions, default to the
        *CIE 1931 2 Degree Standard Observer*.
    illuminant
        Illuminant spectral distribution, default to *CIE Illuminant E*.
    k
        Normalisation constant :math:`k`. For reflecting or transmitting object
        colours, :math:`k` is chosen so that :math:`Y = 100` for objects for
        which the spectral reflectance factor :math:`R(\\lambda)` of the object
        colour or the spectral transmittance factor :math:`\\tau(\\lambda)` of
        the object is equal to unity for all wavelengths. For self-luminous
        objects and illuminants, the constants :math:`k` is usually chosen on
        the grounds of convenience. If, however, in the CIE 1931 standard
        colorimetric system, the :math:`Y` value is required to be numerically
        equal to the absolute value of a photometric quantity, the constant,
        :math:`k`, must be put equal to the numerical value of :math:`K_m`, the
        maximum spectral luminous efficacy (which is equal to
        683 :math:`lm\\cdot W^{-1}`) and :math:`\\Phi_\\lambda(\\lambda)` must
        be the spectral concentration of the radiometric quantity corresponding
        to the photometric quantity required.
    shape
        Spectral shape of the spectral distribution, ``cmfs`` and
        ``illuminant`` will be aligned to it if ``sd`` is an `ArrayLike`.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ* tristimulus values.

    Notes
    -----
    +-----------+-----------------------+---------------+
    | **Range** | **Scale - Reference** | **Scale - 1** |
    +===========+=======================+===============+
    | ``XYZ``   | [0, 100]              | [0, 1]        |
    +-----------+-----------------------+---------------+

    -   The code path using the `ArrayLike` spectral distribution produces
        results different to the code path using a
        :class:`colour.SpectralDistribution` class instance: the former
        favours execution speed by aligning the colour matching functions and
        illuminant to the given spectral shape while the latter favours
        precision by aligning the spectral distribution to the colour matching
        functions.

    References
    ----------
    :cite:`Wyszecki2000bf`

    Examples
    --------
    >>> from colour import MSDS_CMFS, SDS_ILLUMINANTS, SpectralDistribution
    >>> cmfs = MSDS_CMFS['CIE 1931 2 Degree Standard Observer']
    >>> illuminant = SDS_ILLUMINANTS['D65']
    >>> shape = SpectralShape(400, 700, 20)
    >>> data = np.array([
    ...     0.0641, 0.0645, 0.0562, 0.0537, 0.0559, 0.0651, 0.0705, 0.0772,
    ...     0.0870, 0.1128, 0.1360, 0.1511, 0.1688, 0.1996, 0.2397, 0.2852
    ... ])
    >>> sd = SpectralDistribution(data, shape)
    >>> sd_to_XYZ_integration(sd, cmfs, illuminant)
    ... # doctest: +ELLIPSIS
    array([ 10.8404805...,   9.6838697...,   6.2115722...])
    >>> sd_to_XYZ_integration(data, cmfs, illuminant, shape=shape)
    ... # doctest: +ELLIPSIS
    array([ 10.8993917...,   9.6986145...,   6.2540301...])

    # The default CMFS are the "CIE 1931 2 Degree Standard Observer", and the
    # default illuminant is "CIE Illuminant E":

    >>> sd_to_XYZ_integration(sd)
    ... # doctest: +ELLIPSIS
    array([ 11.7786939...,   9.9583972...,   5.7371816...])
    """

    # NOTE: The "illuminant" argument is reshaped by the
    # `handle_spectral_arguments` definition, but, in this case, it is not
    # desirable as we want to reshape it according to the final "shape" which
    # is only available after the subsequent if/else block thus we are careful
    # not unpacking over it.
    if illuminant is None:
        cmfs, illuminant = handle_spectral_arguments(
            cmfs, illuminant, illuminant_default="E"
        )
    else:
        cmfs, _illuminant = handle_spectral_arguments(
            cmfs, illuminant, illuminant_default="E"
        )

    if isinstance(sd, (SpectralDistribution, MultiSpectralDistributions)):
        shape = cmfs.shape

        if sd.shape != shape:
            runtime_warning(
                f'Aligning "{sd.name}" spectral data shape to "{shape}".'
            )

            sd = (
                reshape_sd(sd, shape)
                if isinstance(sd, SpectralDistribution)
                else reshape_msds(sd, shape)
            )

        R = np.transpose(sd.values)
        shape_R = R.shape
        wl_c_r = R.shape[-1]
    else:
        attest(
            shape is not None,
            "A spectral shape must be explicitly passed with a spectral data "
            "array!",
        )

        shape = cast(SpectralShape, shape)

        R = as_float_array(sd)
        shape_R = R.shape
        wl_c_r = R.shape[-1]
        wl_c = len(shape.range())

        attest(
            wl_c_r == wl_c,
            f"Spectral data array with {wl_c_r} wavelengths is not compatible "
            f"with spectral shape with {wl_c} wavelengths!",
        )

        if cmfs.shape != shape:
            runtime_warning(f'Aligning "{cmfs.name}" cmfs shape to "{shape}".')
            # pylint: disable=E1102
            cmfs = reshape_msds(cmfs, shape)

    if illuminant.shape != shape:
        runtime_warning(
            f'Aligning "{illuminant.name}" illuminant shape to "{shape}".'
        )
        illuminant = reshape_sd(illuminant, shape)

    XYZ_b = cmfs.values
    S = illuminant.values
    R = np.reshape(R, (-1, wl_c_r))

    d_w = cmfs.shape.interval

    k = cast(Number, optional(k, 100 / (np.sum(XYZ_b[..., 1] * S) * d_w)))

    XYZ = k * np.dot(R * S, XYZ_b) * d_w

    return from_range_100(np.reshape(XYZ, list(shape_R[:-1]) + [3]))


def sd_to_XYZ_tristimulus_weighting_factors_ASTME308(
    sd: SpectralDistribution,
    cmfs: Optional[MultiSpectralDistributions] = None,
    illuminant: Optional[SpectralDistribution] = None,
    k: Optional[Number] = None,
) -> NDArray:
    """
    Convert given spectral distribution to *CIE XYZ* tristimulus values
    using given colour matching functions and illuminant using a table of
    tristimulus weighting factors according to practise *ASTM E308-15* method.

    Parameters
    ----------
    sd
        Spectral distribution.
    cmfs
        Standard observer colour matching functions, default to the
        *CIE 1931 2 Degree Standard Observer*.
    illuminant
        Illuminant spectral distribution, default to *CIE Illuminant E*.
    k
        Normalisation constant :math:`k`. For reflecting or transmitting object
        colours, :math:`k` is chosen so that :math:`Y = 100` for objects for
        which the spectral reflectance factor :math:`R(\\lambda)` of the object
        colour or the spectral transmittance factor :math:`\\tau(\\lambda)` of
        the object is equal to unity for all wavelengths. For self-luminous
        objects and illuminants, the constants :math:`k` is usually chosen on
        the grounds of convenience. If, however, in the CIE 1931 standard
        colorimetric system, the :math:`Y` value is required to be numerically
        equal to the absolute value of a photometric quantity, the constant,
        :math:`k`, must be put equal to the numerical value of :math:`K_m`, the
        maximum spectral luminous efficacy (which is equal to
        683 :math:`lm\\cdot W^{-1}`) and :math:`\\Phi_\\lambda(\\lambda)` must
        be the spectral concentration of the radiometric quantity corresponding
        to the photometric quantity required.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ* tristimulus values.

    Notes
    -----
    +-----------+-----------------------+---------------+
    | **Range** | **Scale - Reference** | **Scale - 1** |
    +===========+=======================+===============+
    | ``XYZ``   | [0, 100]              | [0, 1]        |
    +-----------+-----------------------+---------------+

    References
    ----------
    :cite:`ASTMInternational2015b`

    Examples
    --------
    >>> from colour import MSDS_CMFS, SDS_ILLUMINANTS, SpectralDistribution
    >>> cmfs = MSDS_CMFS['CIE 1931 2 Degree Standard Observer']
    >>> illuminant = SDS_ILLUMINANTS['D65']
    >>> shape = SpectralShape(400, 700, 20)
    >>> data = np.array([
    ...     0.0641, 0.0645, 0.0562, 0.0537, 0.0559, 0.0651, 0.0705, 0.0772,
    ...     0.0870, 0.1128, 0.1360, 0.1511, 0.1688, 0.1996, 0.2397, 0.2852
    ... ])
    >>> sd = SpectralDistribution(data, shape)
    >>> sd_to_XYZ_tristimulus_weighting_factors_ASTME308(
    ...     sd, cmfs, illuminant)  # doctest: +ELLIPSIS
    array([ 10.8405832...,   9.6844909...,   6.2155622...])

    # The default CMFS are the "CIE 1931 2 Degree Standard Observer", and the
    # default illuminant is "CIE Illuminant E":

    >>> sd_to_XYZ_tristimulus_weighting_factors_ASTME308(sd)
    ... # doctest: +ELLIPSIS
    array([ 11.7786111...,   9.9589055...,   5.7403205...])
    """

    cmfs, illuminant = handle_spectral_arguments(
        cmfs,
        illuminant,
        "CIE 1931 2 Degree Standard Observer",
        "E",
        SPECTRAL_SHAPE_ASTME308,
    )

    if cmfs.shape.interval != 1:
        runtime_warning(f'Interpolating "{cmfs.name}" cmfs to 1nm interval.')
        # pylint: disable=E1102
        cmfs = reshape_msds(
            cmfs,
            SpectralShape(cmfs.shape.start, cmfs.shape.end, 1),
            "Interpolate",
        )

    if illuminant.shape != cmfs.shape:
        runtime_warning(
            f'Aligning "{illuminant.name}" illuminant shape to "{cmfs.name}" '
            f"colour matching functions shape."
        )
        illuminant = reshape_sd(illuminant, cmfs.shape)

    if sd.shape.boundaries != cmfs.shape.boundaries:
        runtime_warning(
            f'Trimming "{illuminant.name}" spectral distribution shape to '
            f'"{cmfs.name}" colour matching functions shape.'
        )
        sd = reshape_sd(sd, cmfs.shape, "Trim")

    W = tristimulus_weighting_factors_ASTME2022(
        cmfs,
        illuminant,
        SpectralShape(cmfs.shape.start, cmfs.shape.end, sd.shape.interval),
        k,
    )
    start_w = cmfs.shape.start
    end_w = cmfs.shape.start + sd.shape.interval * (W.shape[0] - 1)
    W = adjust_tristimulus_weighting_factors_ASTME308(
        W, SpectralShape(start_w, end_w, sd.shape.interval), sd.shape
    )
    R = sd.values

    XYZ = np.sum(W * R[..., np.newaxis], axis=0)

    return from_range_100(XYZ)


def sd_to_XYZ_ASTME308(
    sd: SpectralDistribution,
    cmfs: Optional[MultiSpectralDistributions] = None,
    illuminant: Optional[SpectralDistribution] = None,
    use_practice_range: Boolean = True,
    mi_5nm_omission_method: Boolean = True,
    mi_20nm_interpolation_method: Boolean = True,
    k: Optional[Number] = None,
) -> NDArray:
    """
    Convert given spectral distribution to *CIE XYZ* tristimulus values using
    given colour matching functions and illuminant according to practise
    *ASTM E308-15* method.

    Parameters
    ----------
    sd
        Spectral distribution.
    cmfs
        Standard observer colour matching functions, default to the
        *CIE 1931 2 Degree Standard Observer*.
    illuminant
        Illuminant spectral distribution, default to *CIE Illuminant E*.
    use_practice_range
        Practise *ASTM E308-15* working wavelengths range is [360, 780],
        if *True* this argument will trim the colour matching functions
        appropriately.
    mi_5nm_omission_method
        5 nm measurement intervals spectral distribution conversion to
        tristimulus values will use a 5 nm version of the colour matching
        functions instead of a table of tristimulus weighting factors.
    mi_20nm_interpolation_method
        20 nm measurement intervals spectral distribution conversion to
        tristimulus values will use a dedicated interpolation method instead
        of a table of tristimulus weighting factors.
    k
        Normalisation constant :math:`k`. For reflecting or transmitting object
        colours, :math:`k` is chosen so that :math:`Y = 100` for objects for
        which the spectral reflectance factor :math:`R(\\lambda)` of the object
        colour or the spectral transmittance factor :math:`\\tau(\\lambda)` of
        the object is equal to unity for all wavelengths. For self-luminous
        objects and illuminants, the constants :math:`k` is usually chosen on
        the grounds of convenience. If, however, in the CIE 1931 standard
        colorimetric system, the :math:`Y` value is required to be numerically
        equal to the absolute value of a photometric quantity, the constant,
        :math:`k`, must be put equal to the numerical value of :math:`K_m`, the
        maximum spectral luminous efficacy (which is equal to
        683 :math:`lm\\cdot W^{-1}`) and :math:`\\Phi_\\lambda(\\lambda)` must
        be the spectral concentration of the radiometric quantity corresponding
        to the photometric quantity required.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ* tristimulus values.

    Notes
    -----
    +-----------+-----------------------+---------------+
    | **Range** | **Scale - Reference** | **Scale - 1** |
    +===========+=======================+===============+
    | ``XYZ``   | [0, 100]              | [0, 1]        |
    +-----------+-----------------------+---------------+

    References
    ----------
    :cite:`ASTMInternational2015b`

    Examples
    --------
    >>> from colour import MSDS_CMFS, SDS_ILLUMINANTS, SpectralDistribution
    >>> cmfs = MSDS_CMFS['CIE 1931 2 Degree Standard Observer']
    >>> illuminant = SDS_ILLUMINANTS['D65']
    >>> shape = SpectralShape(400, 700, 20)
    >>> data = np.array([
    ...     0.0641, 0.0645, 0.0562, 0.0537, 0.0559, 0.0651, 0.0705, 0.0772,
    ...     0.0870, 0.1128, 0.1360, 0.1511, 0.1688, 0.1996, 0.2397, 0.2852
    ... ])
    >>> sd = SpectralDistribution(data, shape)
    >>> sd_to_XYZ_ASTME308(sd, cmfs, illuminant)
    ... # doctest: +ELLIPSIS
    array([ 10.8401953...,   9.6841740...,   6.2158913...])

    # The default CMFS are the "CIE 1931 2 Degree Standard Observer", and the
    # default illuminant is "CIE Illuminant E":

    >>> sd_to_XYZ_ASTME308(sd)
    ... # doctest: +ELLIPSIS
    array([ 11.7781589...,   9.9585580...,   5.7408602...])
    """

    cmfs, illuminant = handle_spectral_arguments(
        cmfs,
        illuminant,
        "CIE 1931 2 Degree Standard Observer",
        "E",
        SPECTRAL_SHAPE_ASTME308,
    )

    if sd.shape.interval not in (1, 5, 10, 20):
        raise ValueError(
            "Tristimulus values conversion from spectral data according to "
            'practise "ASTM E308-15" should be performed on spectral data '
            "with measurement interval of 1, 5, 10 or 20nm!"
        )

    if use_practice_range:
        # pylint: disable=E1102
        cmfs = reshape_msds(cmfs, SPECTRAL_SHAPE_ASTME308, "Trim")

    method = sd_to_XYZ_tristimulus_weighting_factors_ASTME308
    if sd.shape.interval == 1:
        method = sd_to_XYZ_integration
    elif sd.shape.interval == 5 and mi_5nm_omission_method:
        if cmfs.shape.interval != 5:
            # pylint: disable=E1102
            cmfs = reshape_msds(
                cmfs,
                SpectralShape(cmfs.shape.start, cmfs.shape.end, 5),
                "Interpolate",
            )
        method = sd_to_XYZ_integration
    elif sd.shape.interval == 20 and mi_20nm_interpolation_method:
        sd = cast(SpectralDistribution, sd.copy())
        if sd.shape.boundaries != cmfs.shape.boundaries:
            runtime_warning(
                f'Trimming "{illuminant.name}" spectral distribution shape to '
                f'"{cmfs.name}" colour matching functions shape.'
            )
            sd.trim(cmfs.shape)

        # Extrapolation of additional 20nm padding intervals.
        sd.align(SpectralShape(sd.shape.start - 20, sd.shape.end + 20, 10))
        for i in range(2):
            sd[sd.wavelengths[i]] = (
                3 * sd.values[i + 2] - 3 * sd.values[i + 4] + sd.values[i + 6]
            )
            i_e = len(sd.domain) - 1 - i
            sd[sd.wavelengths[i_e]] = (
                sd.values[i_e - 6]
                - 3 * sd.values[i_e - 4]
                + 3 * sd.values[i_e - 2]
            )

        # Interpolating every odd numbered values.
        # TODO: Investigate code vectorisation.
        for i in range(3, len(sd.domain) - 3, 2):
            sd[sd.wavelengths[i]] = (
                -0.0625 * sd.values[i - 3]
                + 0.5625 * sd.values[i - 1]
                + 0.5625 * sd.values[i + 1]
                - 0.0625 * sd.values[i + 3]
            )

        # Discarding the additional 20nm padding intervals.
        sd.trim(SpectralShape(sd.shape.start + 20, sd.shape.end - 20, 10))

    XYZ = method(sd, cmfs, illuminant, k=k)

    return XYZ


SD_TO_XYZ_METHODS = CaseInsensitiveMapping(
    {"ASTM E308": sd_to_XYZ_ASTME308, "Integration": sd_to_XYZ_integration}
)
SD_TO_XYZ_METHODS.__doc__ = """
Supported spectral distribution to *CIE XYZ* tristimulus values conversion
methods.

References
----------
:cite:`ASTMInternational2011a`, :cite:`ASTMInternational2015b`,
:cite:`Wyszecki2000bf`

Aliases:

-   'astm2015': 'ASTM E308'
"""
SD_TO_XYZ_METHODS["astm2015"] = SD_TO_XYZ_METHODS["ASTM E308"]


def sd_to_XYZ(
    sd: Union[ArrayLike, SpectralDistribution, MultiSpectralDistributions],
    cmfs: Optional[MultiSpectralDistributions] = None,
    illuminant: Optional[SpectralDistribution] = None,
    k: Optional[Number] = None,
    method: Union[Literal["ASTM E308", "Integration"], str] = "ASTM E308",
    **kwargs: Any,
) -> NDArray:
    """
    Convert given spectral distribution to *CIE XYZ* tristimulus values using
    given colour matching functions, illuminant and method.

    If ``method`` is *Integration*, the spectral distribution can be either a
    :class:`colour.SpectralDistribution` class instance or an `ArrayLike` in
    which case the ``shape`` must be passed.

    Parameters
    ----------
    sd
        Spectral distribution, if an `ArrayLike` and ``method`` is
        *Integration* the wavelengths are expected to be in the last axis, e.g.
        for a spectral array with 77 bins, ``sd`` shape could be (77, ) or
        (1, 77).
    cmfs
        Standard observer colour matching functions, default to the
        *CIE 1931 2 Degree Standard Observer*.
    illuminant
        Illuminant spectral distribution, default to *CIE Illuminant E*.
    k
        Normalisation constant :math:`k`. For reflecting or transmitting object
        colours, :math:`k` is chosen so that :math:`Y = 100` for objects for
        which the spectral reflectance factor :math:`R(\\lambda)` of the object
        colour or the spectral transmittance factor :math:`\\tau(\\lambda)` of
        the object is equal to unity for all wavelengths. For self-luminous
        objects and illuminants, the constants :math:`k` is usually chosen on
        the grounds of convenience. If, however, in the CIE 1931 standard
        colorimetric system, the :math:`Y` value is required to be numerically
        equal to the absolute value of a photometric quantity, the constant,
        :math:`k`, must be put equal to the numerical value of :math:`K_m`, the
        maximum spectral luminous efficacy (which is equal to
        683 :math:`lm\\cdot W^{-1}`) and :math:`\\Phi_\\lambda(\\lambda)` must
        be the spectral concentration of the radiometric quantity corresponding
        to the photometric quantity required.
    method
        Computation method.

    Other Parameters
    ----------------
    mi_5nm_omission_method
        {:func:`colour.colorimetry.sd_to_XYZ_ASTME308`},
        5 nm measurement intervals spectral distribution conversion to
        tristimulus values will use a 5 nm version of the colour matching
        functions instead of a table of tristimulus weighting factors.
    mi_20nm_interpolation_method
        {:func:`colour.colorimetry.sd_to_XYZ_ASTME308`},
        20 nm measurement intervals spectral distribution conversion to
        tristimulus values will use a dedicated interpolation method instead
        of a table of tristimulus weighting factors.
    shape
        Spectral shape of the spectral distribution, ``cmfs`` and
        ``illuminant`` will be aligned to it if ``sd`` is an `ArrayLike`.
    use_practice_range
        {:func:`colour.colorimetry.sd_to_XYZ_ASTME308`},
        Practise *ASTM E308-15* working wavelengths range is [360, 780],
        if *True* this argument will trim the colour matching functions
        appropriately.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ* tristimulus values.

    Notes
    -----
    +-----------+-----------------------+---------------+
    | **Range** | **Scale - Reference** | **Scale - 1** |
    +===========+=======================+===============+
    | ``XYZ``   | [0, 100]              | [0, 1]        |
    +-----------+-----------------------+---------------+

    -   The code path using the `ArrayLike` spectral distribution produces
        results different to the code path using a
        :class:`colour.SpectralDistribution` class instance: the former
        favours execution speed by aligning the colour matching functions and
        illuminant to the given spectral shape while the latter favours
        precision by aligning the spectral distribution to the colour matching
        functions.

    References
    ----------
    :cite:`ASTMInternational2011a`, :cite:`ASTMInternational2015b`,
    :cite:`Wyszecki2000bf`

    Examples
    --------
    >>> import numpy as np
    >>> from colour import (
    ...     MSDS_CMFS, SDS_ILLUMINANTS, SpectralDistribution, SpectralShape)
    >>> cmfs = MSDS_CMFS['CIE 1931 2 Degree Standard Observer']
    >>> illuminant = SDS_ILLUMINANTS['D65']
    >>> shape = SpectralShape(400, 700, 20)
    >>> data = np.array([
    ...     0.0641, 0.0645, 0.0562, 0.0537, 0.0559, 0.0651, 0.0705, 0.0772,
    ...     0.0870, 0.1128, 0.1360, 0.1511, 0.1688, 0.1996, 0.2397, 0.2852
    ... ])
    >>> sd = SpectralDistribution(data, shape)
    >>> sd_to_XYZ(sd, cmfs, illuminant)
    ... # doctest: +ELLIPSIS
    array([ 10.8401953...,   9.6841740...,   6.2158913...])
    >>> sd_to_XYZ(sd, cmfs, illuminant, use_practice_range=False)
    ... # doctest: +ELLIPSIS
    array([ 10.8402774...,   9.6841967...,   6.2158838...])
    >>> sd_to_XYZ(sd, cmfs, illuminant, method='Integration')
    ... # doctest: +ELLIPSIS
    array([ 10.8404805...,   9.6838697...,   6.2115722...])
    >>> sd_to_XYZ(data, cmfs, illuminant, method='Integration', shape=shape)
    ... # doctest: +ELLIPSIS
    array([ 10.8993917...,   9.6986145...,   6.2540301...])

    # The default CMFS are the "CIE 1931 2 Degree Standard Observer", and the
    # default illuminant is "CIE Illuminant E":

    >>> sd_to_XYZ(sd)
    ... # doctest: +ELLIPSIS
    array([ 11.7781589...,   9.9585580...,   5.7408602...])
    """

    cmfs, illuminant = handle_spectral_arguments(
        cmfs, illuminant, illuminant_default="E"
    )

    method = validate_method(method, SD_TO_XYZ_METHODS)

    global _CACHE_SD_TO_XYZ

    hash_key = tuple(
        hash(arg)
        for arg in [
            sd
            if isinstance(
                sd, (SpectralDistribution, MultiSpectralDistributions)
            )
            else sd.tobytes(),  # type: ignore[union-attr]
            cmfs,
            illuminant,
            k,
            method,
            tuple(kwargs.items()),
            get_domain_range_scale(),
        ]
    )
    if hash_key in _CACHE_SD_TO_XYZ:
        return np.copy(_CACHE_SD_TO_XYZ[hash_key])

    function = SD_TO_XYZ_METHODS[method]

    XYZ = function(
        sd, cmfs, illuminant, k=k, **filter_kwargs(function, **kwargs)
    )

    _CACHE_SD_TO_XYZ[hash_key] = np.copy(XYZ)

    return XYZ


def msds_to_XYZ_integration(
    msds: Union[ArrayLike, SpectralDistribution, MultiSpectralDistributions],
    cmfs: Optional[MultiSpectralDistributions] = None,
    illuminant: Optional[SpectralDistribution] = None,
    k: Optional[Number] = None,
    shape: Optional[SpectralShape] = None,
) -> NDArray:
    """
    Convert given multi-spectral distributions to *CIE XYZ* tristimulus values
    using given colour matching functions and illuminant.

    The multi-spectral distributions can be either a
    :class:`colour.MultiSpectralDistributions` class instance or an
    `ArrayLike` in which case the ``shape`` must be passed.

    Parameters
    ----------
    msds
        Multi-spectral distributions, if an `ArrayLike` the wavelengths are
        expected to be in the last axis, e.g. for a 512x384 multi-spectral
        image with 77 bins, ``msds`` shape should be (384, 512, 77).
    cmfs
        Standard observer colour matching functions, default to the
        *CIE 1931 2 Degree Standard Observer*.
    illuminant
        Illuminant spectral distribution, default to *CIE Illuminant E*.
    k
        Normalisation constant :math:`k`. For reflecting or transmitting object
        colours, :math:`k` is chosen so that :math:`Y = 100` for objects for
        which the spectral reflectance factor :math:`R(\\lambda)` of the object
        colour or the spectral transmittance factor :math:`\\tau(\\lambda)` of
        the object is equal to unity for all wavelengths. For self-luminous
        objects and illuminants, the constants :math:`k` is usually chosen on
        the grounds of convenience. If, however, in the CIE 1931 standard
        colorimetric system, the :math:`Y` value is required to be numerically
        equal to the absolute value of a photometric quantity, the constant,
        :math:`k`, must be put equal to the numerical value of :math:`K_m`, the
        maximum spectral luminous efficacy (which is equal to
        683 :math:`lm\\cdot W^{-1}`) and :math:`\\Phi_\\lambda(\\lambda)` must
        be the spectral concentration of the radiometric quantity corresponding
        to the photometric quantity required.
    shape
        Spectral shape of the multi-spectral distributions, ``cmfs`` and
        ``illuminant`` will be aligned to it if ``msds`` is an `ArrayLike`.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ* tristimulus values, for a 512x384 multi-spectral image with
        77 bins, the output shape will be (384, 512, 3).

    Notes
    -----
    +-----------+-----------------------+---------------+
    | **Range** | **Scale - Reference** | **Scale - 1** |
    +===========+=======================+===============+
    | ``XYZ``   | [0, 100]              | [0, 1]        |
    +-----------+-----------------------+---------------+

    -   The code path using the `ArrayLike` multi-spectral distributions
        produces results different to the code path using a
        :class:`colour.MultiSpectralDistributions` class instance: the former
        favours execution speed by aligning the colour matching functions and
        illuminant to the given spectral shape while the latter favours
        precision by aligning the multi-spectral distributions to the colour
        matching functions.

    References
    ----------
    :cite:`Wyszecki2000bf`

    Examples
    --------
    >>> from colour import MSDS_CMFS, SDS_ILLUMINANTS
    >>> cmfs = MSDS_CMFS['CIE 1931 2 Degree Standard Observer']
    >>> illuminant = SDS_ILLUMINANTS['D65']
    >>> shape = SpectralShape(400, 700, 60)
    >>> data = np.array([
    ...     [0.0137, 0.0159, 0.0096, 0.0111, 0.0179, 0.1057, 0.0433,
    ...      0.0258, 0.0248, 0.0186, 0.0310, 0.0473],
    ...     [0.0913, 0.3145, 0.2582, 0.0709, 0.2971, 0.4620, 0.2683,
    ...      0.0831, 0.1203, 0.1292, 0.1682, 0.3221],
    ...     [0.0152, 0.0842, 0.4139, 0.0220, 0.5630, 0.1918, 0.2373,
    ...      0.0430, 0.0054, 0.0079, 0.3719, 0.2268],
    ...     [0.0281, 0.0907, 0.2228, 0.1249, 0.2375, 0.5625, 0.0518,
    ...      0.3230, 0.0065, 0.4006, 0.0861, 0.3161],
    ...     [0.1918, 0.7103, 0.0041, 0.1817, 0.0024, 0.4209, 0.0118,
    ...      0.2302, 0.1860, 0.9404, 0.0041, 0.1124],
    ...     [0.0430, 0.0437, 0.3744, 0.0020, 0.5819, 0.0027, 0.0823,
    ...      0.0081, 0.3625, 0.3213, 0.7849, 0.0024],
    ... ])
    >>> msds = MultiSpectralDistributions(data, shape)
    >>> msds_to_XYZ_integration(msds, cmfs, illuminant)
    ... # doctest: +ELLIPSIS
    array([[  7.5029704...,   3.9487844...,   8.4034669...],
           [ 26.9259681...,  15.0724609...,  28.7057807...],
           [ 16.7032188...,  28.2172346...,  25.6455984...],
           [ 11.5767013...,   8.6400993...,   6.5768406...],
           [ 18.7314793...,  35.0750364...,  30.1457266...],
           [ 45.1656756...,  39.6136917...,  43.6783499...],
           [  8.1755696...,  13.0934177...,  25.9420944...],
           [ 22.4676286...,  19.3099080...,   7.9637549...],
           [  6.5781241...,   2.5255349...,  11.0930768...],
           [ 43.9147364...,  27.9803924...,  11.7292655...],
           [  8.5365923...,  19.7030166...,  17.7050933...],
           [ 23.9088250...,  26.2129529...,  30.6763148...]])
    >>> data = np.reshape(data, (2, 6, 6))
    >>> msds_to_XYZ_integration(data, cmfs, illuminant, shape=shape)
    ... # doctest: +ELLIPSIS
    array([[[  1.3104332...,   1.1377026...,   1.8267926...],
            [  2.1875548...,   2.2510619...,   3.0721540...],
            [ 16.8714661...,  17.7063715...,  35.8709902...],
            [ 12.1648722...,  12.7222194...,  10.4880888...],
            [ 16.0419431...,  23.0985768...,  11.1479902...],
            [  9.2391014...,   3.8301575...,   5.4703803...]],
    <BLANKLINE>
           [[ 13.8734231...,  17.3942194...,  11.0364103...],
            [ 27.7096381...,  20.8626722...,  35.5581690...],
            [ 22.7886687...,  11.4769218...,  78.3300659...],
            [ 51.1284864...,  52.2463568...,  26.1483754...],
            [ 14.4749229...,  20.5011495...,   6.6228107...],
            [ 33.6001365...,  36.3242617...,   2.8254217...]]])

    # The default CMFS are the "CIE 1931 2 Degree Standard Observer", and the
    # default illuminant is "CIE Illuminant E":

    >>> msds_to_XYZ_integration(msds)
    ... # doctest: +ELLIPSIS
    array([[  8.2415862...,   4.2543993...,   7.6100842...],
           [ 29.6144619...,  16.1158465...,  25.9015472...],
           [ 16.6799560...,  27.2350547...,  22.9413337...],
           [ 12.5597688...,   9.0667136...,   5.9670327...],
           [ 18.5804689...,  33.6618109...,  26.9249733...],
           [ 47.7113308...,  40.4573249...,  39.6439145...],
           [  7.830207 ...,  12.3689624...,  23.3742655...],
           [ 24.1695370...,  20.0629815...,   7.2718670...],
           [  7.2333751...,   2.7982097...,  10.0688374...],
           [ 48.7358074...,  30.2417164...,  10.6753233...],
           [  8.3231013...,  18.6791507...,  15.8228184...],
           [ 24.6452277...,  26.0809382...,  27.7106399...]])
    """

    return sd_to_XYZ_integration(msds, cmfs, illuminant, k, shape)


def msds_to_XYZ_ASTME308(
    msds: MultiSpectralDistributions,
    cmfs: Optional[MultiSpectralDistributions] = None,
    illuminant: Optional[SpectralDistribution] = None,
    use_practice_range: Boolean = True,
    mi_5nm_omission_method: Boolean = True,
    mi_20nm_interpolation_method: Boolean = True,
    k: Optional[Number] = None,
) -> NDArray:
    """
    Convert given multi-spectral distributions to *CIE XYZ* tristimulus values
    using given colour matching functions and illuminant according to practise
    *ASTM E308-15* method.

    Parameters
    ----------
    msds
        Multi-spectral distributions.
    cmfs
        Standard observer colour matching functions, default to the
        *CIE 1931 2 Degree Standard Observer*.
    illuminant
        Illuminant spectral distribution, default to *CIE Illuminant E*.
    use_practice_range
        Practise *ASTM E308-15* working wavelengths range is [360, 780],
        if *True* this argument will trim the colour matching functions
        appropriately.
    mi_5nm_omission_method
        5 nm measurement intervals multi-spectral distributions conversion to
        tristimulus values will use a 5 nm version of the colour matching
        functions instead of a table of tristimulus weighting factors.
    mi_20nm_interpolation_method
        20 nm measurement intervals multi-spectral distributions conversion to
        tristimulus values will use a dedicated interpolation method instead
        of a table of tristimulus weighting factors.
    k
        Normalisation constant :math:`k`. For reflecting or transmitting object
        colours, :math:`k` is chosen so that :math:`Y = 100` for objects for
        which the spectral reflectance factor :math:`R(\\lambda)` of the object
        colour or the spectral transmittance factor :math:`\\tau(\\lambda)` of
        the object is equal to unity for all wavelengths. For self-luminous
        objects and illuminants, the constants :math:`k` is usually chosen on
        the grounds of convenience. If, however, in the CIE 1931 standard
        colorimetric system, the :math:`Y` value is required to be numerically
        equal to the absolute value of a photometric quantity, the constant,
        :math:`k`, must be put equal to the numerical value of :math:`K_m`, the
        maximum spectral luminous efficacy (which is equal to
        683 :math:`lm\\cdot W^{-1}`) and :math:`\\Phi_\\lambda(\\lambda)` must
        be the spectral concentration of the radiometric quantity corresponding
        to the photometric quantity required.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ* tristimulus values.

    Notes
    -----
    +-----------+-----------------------+---------------+
    | **Range** | **Scale - Reference** | **Scale - 1** |
    +===========+=======================+===============+
    | ``XYZ``   | [0, 100]              | [0, 1]        |
    +-----------+-----------------------+---------------+

    -   The code path using the `ArrayLike` multi-spectral distributions
        produces results different to the code path using a
        :class:`colour.MultiSpectralDistributions` class instance: the former
        favours execution speed by aligning the colour matching functions and
        illuminant to the given spectral shape while the latter favours
        precision by aligning the multi-spectral distributions to the colour
        matching functions.

    References
    ----------
    :cite:`Wyszecki2000bf`

    Examples
    --------
    >>> from colour import MSDS_CMFS, SDS_ILLUMINANTS
    >>> cmfs = MSDS_CMFS['CIE 1931 2 Degree Standard Observer']
    >>> illuminant = SDS_ILLUMINANTS['D65']
    >>> shape = SpectralShape(400, 700, 60)
    >>> data = np.array([
    ...     [0.0137, 0.0159, 0.0096, 0.0111, 0.0179, 0.1057, 0.0433,
    ...      0.0258, 0.0248, 0.0186, 0.0310, 0.0473],
    ...     [0.0913, 0.3145, 0.2582, 0.0709, 0.2971, 0.4620, 0.2683,
    ...      0.0831, 0.1203, 0.1292, 0.1682, 0.3221],
    ...     [0.0152, 0.0842, 0.4139, 0.0220, 0.5630, 0.1918, 0.2373,
    ...      0.0430, 0.0054, 0.0079, 0.3719, 0.2268],
    ...     [0.0281, 0.0907, 0.2228, 0.1249, 0.2375, 0.5625, 0.0518,
    ...      0.3230, 0.0065, 0.4006, 0.0861, 0.3161],
    ...     [0.1918, 0.7103, 0.0041, 0.1817, 0.0024, 0.4209, 0.0118,
    ...      0.2302, 0.1860, 0.9404, 0.0041, 0.1124],
    ...     [0.0430, 0.0437, 0.3744, 0.0020, 0.5819, 0.0027, 0.0823,
    ...      0.0081, 0.3625, 0.3213, 0.7849, 0.0024],
    ... ])
    >>> msds = MultiSpectralDistributions(data, shape)
    >>> msds = msds.align(SpectralShape(400, 700, 20))
    >>> msds_to_XYZ_ASTME308(msds, cmfs, illuminant)
    ... # doctest: +ELLIPSIS
    array([[  7.5052758...,   3.9557516...,   8.38929  ...],
           [ 26.9408494...,  15.0987746...,  28.6631260...],
           [ 16.7047370...,  28.2089815...,  25.6556751...],
           [ 11.5711808...,   8.6445071...,   6.5587827...],
           [ 18.7428858...,  35.0626352...,  30.1778517...],
           [ 45.1224886...,  39.6238997...,  43.5813345...],
           [  8.1786985...,  13.0950215...,  25.9326459...],
           [ 22.4462888...,  19.3115133...,   7.9304333...],
           [  6.5764361...,   2.5305945...,  11.07253  ...],
           [ 43.9113380...,  28.0003541...,  11.6852531...],
           [  8.5496209...,  19.6913570...,  17.7400079...],
           [ 23.8866733...,  26.2147704...,  30.6297684...]])

    # The default CMFS are the "CIE 1931 2 Degree Standard Observer", and the
    # default illuminant is "CIE Illuminant E":

    >>> msds_to_XYZ_ASTME308(msds)
    ... # doctest: +ELLIPSIS
    array([[  8.2439318...,   4.2617641...,   7.5977409...],
           [ 29.6290771...,  16.1443076...,  25.8640484...],
           [ 16.6819067...,  27.2271403...,  22.9490590...],
           [ 12.5543694...,   9.0705685...,   5.9516323...],
           [ 18.5921357...,  33.6508573...,  26.9511144...],
           [ 47.6698072...,  40.4630866...,  39.5612904...],
           [  7.8336896...,  12.3711768...,  23.3654245...],
           [ 24.1486630...,  20.0621956...,   7.2438655...],
           [  7.2323703...,   2.8033217...,  10.0510790...],
           [ 48.7322793...,  30.2614779...,  10.6377135...],
           [  8.3365770...,  18.6690888...,  15.8517212...],
           [ 24.6240657...,  26.0805317...,  27.6706915...]])
    """

    cmfs, illuminant = handle_spectral_arguments(
        cmfs,
        illuminant,
        "CIE 1931 2 Degree Standard Observer",
        "E",
        SPECTRAL_SHAPE_ASTME308,
    )

    if isinstance(msds, MultiSpectralDistributions):
        return as_float_array(
            [
                sd_to_XYZ_ASTME308(
                    sd,
                    cmfs,
                    illuminant,
                    use_practice_range,
                    mi_5nm_omission_method,
                    mi_20nm_interpolation_method,
                    k,
                )
                for sd in msds.to_sds()
            ]
        )
    else:
        raise ValueError(
            '"ASTM E308-15" method does not support "ArrayLike" '
            "multi-spectral distributions!"
        )


MSDS_TO_XYZ_METHODS = CaseInsensitiveMapping(
    {"ASTM E308": msds_to_XYZ_ASTME308, "Integration": msds_to_XYZ_integration}
)
MSDS_TO_XYZ_METHODS.__doc__ = """
Supported multi-spectral array to *CIE XYZ* tristimulus values conversion
methods.

References
----------
:cite:`ASTMInternational2011a`, :cite:`ASTMInternational2015b`,
:cite:`Wyszecki2000bf`

Aliases:

-   'astm2015': 'ASTM E308'
"""
MSDS_TO_XYZ_METHODS["astm2015"] = MSDS_TO_XYZ_METHODS["ASTM E308"]


def msds_to_XYZ(
    msds: Union[ArrayLike, SpectralDistribution, MultiSpectralDistributions],
    cmfs: Optional[MultiSpectralDistributions] = None,
    illuminant: Optional[SpectralDistribution] = None,
    k: Optional[Number] = None,
    method: Union[Literal["ASTM E308", "Integration"], str] = "ASTM E308",
    **kwargs: Any,
) -> NDArray:
    """
    Convert given multi-spectral distributions to *CIE XYZ* tristimulus values
    using given colour matching functions and illuminant. For the *Integration*
    method, the multi-spectral distributions can be either a
    :class:`colour.MultiSpectralDistributions` class instance or an
    `ArrayLike` in which case the ``shape`` must be passed.

    Parameters
    ----------
    msds
        Multi-spectral distributions, if an `ArrayLike` the wavelengths are
        expected to be in the last axis, e.g. for a 512x384 multi-spectral
        image with 77 bins, ``msds`` shape should be (384, 512, 77).
    cmfs
        Standard observer colour matching functions, default to the
        *CIE 1931 2 Degree Standard Observer*.
    illuminant
        Illuminant spectral distribution, default to *CIE Illuminant E*.
    k
        Normalisation constant :math:`k`. For reflecting or transmitting object
        colours, :math:`k` is chosen so that :math:`Y = 100` for objects for
        which the spectral reflectance factor :math:`R(\\lambda)` of the object
        colour or the spectral transmittance factor :math:`\\tau(\\lambda)` of
        the object is equal to unity for all wavelengths. For self-luminous
        objects and illuminants, the constants :math:`k` is usually chosen on
        the grounds of convenience. If, however, in the CIE 1931 standard
        colorimetric system, the :math:`Y` value is required to be numerically
        equal to the absolute value of a photometric quantity, the constant,
        :math:`k`, must be put equal to the numerical value of :math:`K_m`, the
        maximum spectral luminous efficacy (which is equal to
        683 :math:`lm\\cdot W^{-1}`) and :math:`\\Phi_\\lambda(\\lambda)` must
        be the spectral concentration of the radiometric quantity corresponding
        to the photometric quantity required.
    method
        Computation method.

    Other Parameters
    ----------------
    mi_5nm_omission_method
        {:func:`colour.colorimetry.msds_to_XYZ_ASTME308`},
        5 nm measurement intervals multi-spectral distributions conversion to
        tristimulus values will use a 5 nm version of the colour matching
        functions instead of a table of tristimulus weighting factors.
    mi_20nm_interpolation_method
        {:func:`colour.colorimetry.msds_to_XYZ_ASTME308`},
        20 nm measurement intervals multi-spectral distributions conversion to
        tristimulus values will use a dedicated interpolation method instead
        of a table of tristimulus weighting factors.
    shape
        {:func:`colour.colorimetry.msds_to_XYZ_integration`},
        Spectral shape of the multi-spectral distributions array :math:`msds`,
        ``cmfs`` and ``illuminant`` will be aligned to it.
    use_practice_range
        {:func:`colour.colorimetry.msds_to_XYZ_ASTME308`},
        Practise *ASTM E308-15* working wavelengths range is [360, 780],
        if *True* this argument will trim the colour matching functions
        appropriately.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ* tristimulus values, for a 512x384 multi-spectral image with
        77 wavelengths, the output shape will be (384, 512, 3).

    Notes
    -----
    +-----------+-----------------------+---------------+
    | **Range** | **Scale - Reference** | **Scale - 1** |
    +===========+=======================+===============+
    | ``XYZ``   | [0, 100]              | [0, 1]        |
    +-----------+-----------------------+---------------+

    -   The code path using the `ArrayLike` multi-spectral distributions
        produces results different to the code path using a
        :class:`colour.MultiSpectralDistributions` class instance: the former
        favours execution speed by aligning the colour matching functions and
        illuminant to the given spectral shape while the latter favours
        precision by aligning the multi-spectral distributions to the colour
        matching functions.

    References
    ----------
    :cite:`ASTMInternational2011a`, :cite:`ASTMInternational2015b`,
    :cite:`Wyszecki2000bf`

    Examples
    --------
    >>> from colour import MSDS_CMFS, SDS_ILLUMINANTS, SpectralDistribution
    >>> cmfs = MSDS_CMFS['CIE 1931 2 Degree Standard Observer']
    >>> illuminant = SDS_ILLUMINANTS['D65']
    >>> shape = SpectralShape(400, 700, 60)
    >>> data = np.array([
    ...     [0.0137, 0.0159, 0.0096, 0.0111, 0.0179, 0.1057, 0.0433,
    ...      0.0258, 0.0248, 0.0186, 0.0310, 0.0473],
    ...     [0.0913, 0.3145, 0.2582, 0.0709, 0.2971, 0.4620, 0.2683,
    ...      0.0831, 0.1203, 0.1292, 0.1682, 0.3221],
    ...     [0.0152, 0.0842, 0.4139, 0.0220, 0.5630, 0.1918, 0.2373,
    ...      0.0430, 0.0054, 0.0079, 0.3719, 0.2268],
    ...     [0.0281, 0.0907, 0.2228, 0.1249, 0.2375, 0.5625, 0.0518,
    ...      0.3230, 0.0065, 0.4006, 0.0861, 0.3161],
    ...     [0.1918, 0.7103, 0.0041, 0.1817, 0.0024, 0.4209, 0.0118,
    ...      0.2302, 0.1860, 0.9404, 0.0041, 0.1124],
    ...     [0.0430, 0.0437, 0.3744, 0.0020, 0.5819, 0.0027, 0.0823,
    ...      0.0081, 0.3625, 0.3213, 0.7849, 0.0024],
    ... ])
    >>> msds = MultiSpectralDistributions(data, shape)
    >>> msds_to_XYZ(msds, cmfs, illuminant, method='Integration')
    ... # doctest: +ELLIPSIS
    array([[  7.5029704...,   3.9487844...,   8.4034669...],
           [ 26.9259681...,  15.0724609...,  28.7057807...],
           [ 16.7032188...,  28.2172346...,  25.6455984...],
           [ 11.5767013...,   8.6400993...,   6.5768406...],
           [ 18.7314793...,  35.0750364...,  30.1457266...],
           [ 45.1656756...,  39.6136917...,  43.6783499...],
           [  8.1755696...,  13.0934177...,  25.9420944...],
           [ 22.4676286...,  19.3099080...,   7.9637549...],
           [  6.5781241...,   2.5255349...,  11.0930768...],
           [ 43.9147364...,  27.9803924...,  11.7292655...],
           [  8.5365923...,  19.7030166...,  17.7050933...],
           [ 23.9088250...,  26.2129529...,  30.6763148...]])
    >>> data = np.reshape(data, (2, 6, 6))
    >>> msds_to_XYZ(data, cmfs, illuminant, method='Integration', shape=shape)
    ... # doctest: +ELLIPSIS
    array([[[  1.3104332...,   1.1377026...,   1.8267926...],
            [  2.1875548...,   2.2510619...,   3.0721540...],
            [ 16.8714661...,  17.7063715...,  35.8709902...],
            [ 12.1648722...,  12.7222194...,  10.4880888...],
            [ 16.0419431...,  23.0985768...,  11.1479902...],
            [  9.2391014...,   3.8301575...,   5.4703803...]],
    <BLANKLINE>
           [[ 13.8734231...,  17.3942194...,  11.0364103...],
            [ 27.7096381...,  20.8626722...,  35.5581690...],
            [ 22.7886687...,  11.4769218...,  78.3300659...],
            [ 51.1284864...,  52.2463568...,  26.1483754...],
            [ 14.4749229...,  20.5011495...,   6.6228107...],
            [ 33.6001365...,  36.3242617...,   2.8254217...]]])

    # The default CMFS are the "CIE 1931 2 Degree Standard Observer", and the
    # default illuminant is "CIE Illuminant E":

    >>> msds_to_XYZ(msds, method='Integration')
    ... # doctest: +ELLIPSIS
    array([[  8.2415862...,   4.2543993...,   7.6100842...],
           [ 29.6144619...,  16.1158465...,  25.9015472...],
           [ 16.6799560...,  27.2350547...,  22.9413337...],
           [ 12.5597688...,   9.0667136...,   5.9670327...],
           [ 18.5804689...,  33.6618109...,  26.9249733...],
           [ 47.7113308...,  40.4573249...,  39.6439145...],
           [  7.830207 ...,  12.3689624...,  23.3742655...],
           [ 24.1695370...,  20.0629815...,   7.2718670...],
           [  7.2333751...,   2.7982097...,  10.0688374...],
           [ 48.7358074...,  30.2417164...,  10.6753233...],
           [  8.3231013...,  18.6791507...,  15.8228184...],
           [ 24.6452277...,  26.0809382...,  27.7106399...]])
    """

    method = validate_method(method, MSDS_TO_XYZ_METHODS)

    function = MSDS_TO_XYZ_METHODS[method]

    return function(
        msds, cmfs, illuminant, k, **filter_kwargs(function, **kwargs)
    )


def wavelength_to_XYZ(
    wavelength: FloatingOrNDArray,
    cmfs: Optional[MultiSpectralDistributions] = None,
) -> NDArray:
    """
    Convert given wavelength :math:`\\lambda` to *CIE XYZ* tristimulus values
    using given colour matching functions.

    If the wavelength :math:`\\lambda` is not available in the colour matching
    function, its value will be calculated according to *CIE 15:2004*
    recommendation: the method developed by *Sprague (1880)* will be used for
    interpolating functions having a uniformly spaced independent variable and
    the *Cubic Spline* method for non-uniformly spaced independent variable.

    Parameters
    ----------
    wavelength
        Wavelength :math:`\\lambda` in nm.
    cmfs
        Standard observer colour matching functions, default to the
        *CIE 1931 2 Degree Standard Observer*.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ* tristimulus values.

    Raises
    ------
    ValueError
        If wavelength :math:`\\lambda` is not contained in the colour matching
        functions domain.

    Notes
    -----
    +-----------+-----------------------+---------------+
    | **Range** | **Scale - Reference** | **Scale - 1** |
    +===========+=======================+===============+
    | ``XYZ``   | [0, 1]                | [0, 1]        |
    +-----------+-----------------------+---------------+

    Examples
    --------
    >>> from colour import MSDS_CMFS
    >>> cmfs = MSDS_CMFS['CIE 1931 2 Degree Standard Observer']
    >>> wavelength_to_XYZ(480, cmfs)  # doctest: +ELLIPSIS
    array([ 0.09564  ,  0.13902  ,  0.8129501...])
    >>> wavelength_to_XYZ(480.5, cmfs)  # doctest: +ELLIPSIS
    array([ 0.0914287...,  0.1418350...,  0.7915726...])
    """

    cmfs, _illuminant = handle_spectral_arguments(cmfs)

    shape = cmfs.shape
    if np.min(wavelength) < shape.start or np.max(wavelength) > shape.end:
        raise ValueError(
            f'"{wavelength}nm" wavelength is not in '
            f'"[{shape.start}, {shape.end}]" domain!'
        )

    XYZ = np.reshape(
        cmfs[np.ravel(wavelength)], as_float_array(wavelength).shape + (3,)
    )

    return XYZ



"""Defines the unit tests for the :mod:`colour.models.cie_uvw` module."""

import numpy as np
import unittest
from itertools import permutations

from colour.models import UVW_to_XYZ, XYZ_to_UVW
from colour.utilities import domain_range_scale, ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestXYZ_to_UVW",
    "TestUVW_to_XYZ",
]


class TestXYZ_to_UVW(unittest.TestCase):
    """
    Define :func:`colour.models.cie_uvw.XYZ_to_UVW` definition unit tests
    methods.
    """

    def test_XYZ_to_UVW(self):
        """Test :func:`colour.models.cie_uvw.XYZ_to_UVW` definition."""

        np.testing.assert_almost_equal(
            XYZ_to_UVW(np.array([0.20654008, 0.12197225, 0.05136952]) * 100),
            np.array([94.55035725, 11.55536523, 40.54757405]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_UVW(np.array([0.14222010, 0.23042768, 0.10495772]) * 100),
            np.array([-36.92762376, 28.90425105, 54.14071478]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_UVW(np.array([0.07818780, 0.06157201, 0.28099326]) * 100),
            np.array([-10.60111550, -41.94580000, 28.82134002]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_UVW(
                np.array([0.20654008, 0.12197225, 0.05136952]) * 100,
                np.array([0.44757, 0.40745]),
            ),
            np.array([63.90676310, -8.11466183, 40.54757405]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_UVW(
                np.array([0.20654008, 0.12197225, 0.05136952]) * 100,
                np.array([0.34570, 0.35850]),
            ),
            np.array([88.56798946, 4.61154385, 40.54757405]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_UVW(
                np.array([0.20654008, 0.12197225, 0.05136952]) * 100,
                np.array([0.34570, 0.35850, 1.00000]),
            ),
            np.array([88.56798946, 4.61154385, 40.54757405]),
            decimal=7,
        )

    def test_n_dimensional_XYZ_to_UVW(self):
        """
        Test :func:`colour.models.cie_uvw.XYZ_to_UVW` definition n-dimensional
        support.
        """

        XYZ = np.array([0.20654008, 0.12197225, 0.05136952]) * 100
        illuminant = np.array([0.31270, 0.32900])
        UVW = XYZ_to_UVW(XYZ, illuminant)

        XYZ = np.tile(XYZ, (6, 1))
        UVW = np.tile(UVW, (6, 1))
        np.testing.assert_almost_equal(
            XYZ_to_UVW(XYZ, illuminant), UVW, decimal=7
        )

        illuminant = np.tile(illuminant, (6, 1))
        np.testing.assert_almost_equal(
            XYZ_to_UVW(XYZ, illuminant), UVW, decimal=7
        )

        XYZ = np.reshape(XYZ, (2, 3, 3))
        illuminant = np.reshape(illuminant, (2, 3, 2))
        UVW = np.reshape(UVW, (2, 3, 3))
        np.testing.assert_almost_equal(
            XYZ_to_UVW(XYZ, illuminant), UVW, decimal=7
        )

    def test_domain_range_scale_XYZ_to_UVW(self):
        """
        Test :func:`colour.models.cie_uvw.XYZ_to_UVW` definition domain and
        range scale support.
        """

        XYZ = np.array([0.20654008, 0.12197225, 0.05136952]) * 100
        illuminant = np.array([0.31270, 0.32900])
        UVW = XYZ_to_UVW(XYZ, illuminant)

        d_r = (("reference", 1), ("1", 0.01), ("100", 1))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    XYZ_to_UVW(XYZ * factor, illuminant),
                    UVW * factor,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_XYZ_to_UVW(self):
        """Test :func:`colour.models.cie_uvw.XYZ_to_UVW` definition nan support."""

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            XYZ = np.array(case)
            illuminant = np.array(case[0:2])
            XYZ_to_UVW(XYZ, illuminant)


class TestUVW_to_XYZ(unittest.TestCase):
    """
    Define :func:`colour.models.cie_uvw.UVW_to_XYZ` definition unit tests
    methods.
    """

    def test_UVW_to_XYZ(self):
        """Test :func:`colour.models.cie_uvw.UVW_to_XYZ` definition."""

        np.testing.assert_almost_equal(
            UVW_to_XYZ(np.array([94.55035725, 11.55536523, 40.54757405])),
            np.array([0.20654008, 0.12197225, 0.05136952]) * 100,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            UVW_to_XYZ(np.array([-36.92762376, 28.90425105, 54.14071478])),
            np.array([0.14222010, 0.23042768, 0.10495772]) * 100,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            UVW_to_XYZ(np.array([-10.60111550, -41.94580000, 28.82134002])),
            np.array([0.07818780, 0.06157201, 0.28099326]) * 100,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            UVW_to_XYZ(
                np.array([63.90676310, -8.11466183, 40.54757405]),
                np.array([0.44757, 0.40745]),
            ),
            np.array([0.20654008, 0.12197225, 0.05136952]) * 100,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            UVW_to_XYZ(
                np.array([88.56798946, 4.61154385, 40.54757405]),
                np.array([0.34570, 0.35850]),
            ),
            np.array([0.20654008, 0.12197225, 0.05136952]) * 100,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            UVW_to_XYZ(
                np.array([88.56798946, 4.61154385, 40.54757405]),
                np.array([0.34570, 0.35850, 1.00000]),
            ),
            np.array([0.20654008, 0.12197225, 0.05136952]) * 100,
            decimal=7,
        )

    def test_n_dimensional_UVW_to_XYZ(self):
        """
        Test :func:`colour.models.cie_uvw.UVW_to_XYZ` definition n-dimensional
        support.
        """

        UVW = np.array([94.55035725, 11.55536523, 40.54757405])
        illuminant = np.array([0.31270, 0.32900])
        XYZ = UVW_to_XYZ(UVW, illuminant)

        XYZ = np.tile(XYZ, (6, 1))
        UVW = np.tile(UVW, (6, 1))
        np.testing.assert_almost_equal(
            UVW_to_XYZ(UVW, illuminant), XYZ, decimal=7
        )

        illuminant = np.tile(illuminant, (6, 1))
        np.testing.assert_almost_equal(
            UVW_to_XYZ(UVW, illuminant), XYZ, decimal=7
        )

        XYZ = np.reshape(XYZ, (2, 3, 3))
        illuminant = np.reshape(illuminant, (2, 3, 2))
        UVW = np.reshape(UVW, (2, 3, 3))
        np.testing.assert_almost_equal(
            UVW_to_XYZ(UVW, illuminant), XYZ, decimal=7
        )

    def test_domain_range_scale_UVW_to_XYZ(self):
        """
        Test :func:`colour.models.cie_uvw.UVW_to_XYZ` definition domain and
        range scale support.
        """

        UVW = np.array([94.55035725, 11.55536523, 40.54757405])
        illuminant = np.array([0.31270, 0.32900])
        XYZ = UVW_to_XYZ(UVW, illuminant)

        d_r = (("reference", 1), ("1", 0.01), ("100", 1))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    UVW_to_XYZ(UVW * factor, illuminant),
                    XYZ * factor,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_UVW_to_XYZ(self):
        """Test :func:`colour.models.cie_uvw.UVW_to_XYZ` definition nan support."""

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            UVW = np.array(case)
            illuminant = np.array(case[0:2])
            UVW_to_XYZ(UVW, illuminant)


if __name__ == "__main__":
    unittest.main()



"""
Hdr-CIELAB Colourspace
======================

Defines the *hdr-CIELAB* colourspace transformations:

-   :attr:`colour.HDR_CIELAB_METHODS`: Supported *hdr-CIELAB* colourspace
    computation methods.
-   :func:`colour.XYZ_to_hdr_CIELab`
-   :func:`colour.hdr_CIELab_to_XYZ`

References
----------
-   :cite:`Fairchild2010` : Fairchild, M. D., & Wyble, D. R. (2010). hdr-CIELAB
    and hdr-IPT: Simple Models for Describing the Color of High-Dynamic-Range
    and Wide-Color-Gamut Images. Proc. of Color and Imaging Conference,
    322-326. ISBN:978-1-62993-215-6
-   :cite:`Fairchild2011` : Fairchild, M. D., & Chen, P. (2011). Brightness,
    lightness, and specifying color in high-dynamic-range scenes and images. In
    S. P. Farnand & F. Gaykema (Eds.), Proc. SPIE 7867, Image Quality and
    System Performance VIII (p. 78670O). doi:10.1117/12.872075
"""

from __future__ import annotations

import numpy as np

from colour.colorimetry import (
    CCS_ILLUMINANTS,
    lightness_Fairchild2010,
    lightness_Fairchild2011,
    luminance_Fairchild2010,
    luminance_Fairchild2011,
)
from colour.hints import (
    ArrayLike,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    Literal,
    NDArray,
    Tuple,
    Union,
)
from colour.models import xy_to_xyY, xyY_to_XYZ
from colour.utilities import (
    as_float_array,
    domain_range_scale,
    from_range_1,
    from_range_100,
    to_domain_1,
    to_domain_100,
    tsplit,
    tstack,
    validate_method,
)
from colour.utilities.documentation import (
    DocstringTuple,
    is_documentation_building,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "HDR_CIELAB_METHODS",
    "exponent_hdr_CIELab",
    "XYZ_to_hdr_CIELab",
    "hdr_CIELab_to_XYZ",
]

HDR_CIELAB_METHODS: Tuple = ("Fairchild 2010", "Fairchild 2011")
if is_documentation_building():  # pragma: no cover
    HDR_CIELAB_METHODS = DocstringTuple(HDR_CIELAB_METHODS)
    HDR_CIELAB_METHODS.__doc__ = """
Supported *hdr-CIELAB* colourspace computation methods.

References
----------
:cite:`Fairchild2010`, :cite:`Fairchild2011`
"""


def exponent_hdr_CIELab(
    Y_s: FloatingOrArrayLike,
    Y_abs: FloatingOrArrayLike,
    method: Union[
        Literal["Fairchild 2011", "Fairchild 2010"], str
    ] = "Fairchild 2011",
) -> FloatingOrNDArray:
    """
    Compute *hdr-CIELAB* colourspace *Lightness* :math:`\\epsilon` exponent
    using *Fairchild and Wyble (2010)* or *Fairchild and Chen (2011)* method.

    Parameters
    ----------
    Y_s
        Relative luminance :math:`Y_s` of the surround.
    Y_abs
        Absolute luminance :math:`Y_{abs}` of the scene diffuse white in
        :math:`cd/m^2`.
    method
        Computation method.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *hdr-CIELAB* colourspace *Lightness* :math:`\\epsilon` exponent.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``Y_s``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    Examples
    --------
    >>> exponent_hdr_CIELab(0.2, 100)  # doctest: +ELLIPSIS
    0.4738510...
    >>> exponent_hdr_CIELab(0.2, 100, method='Fairchild 2010')
    ... # doctest: +ELLIPSIS
    1.8360198...
    """

    Y_s = to_domain_1(Y_s)
    Y_abs = as_float_array(Y_abs)
    method = validate_method(method, HDR_CIELAB_METHODS)

    if method == "fairchild 2010":
        epsilon = 1.50
    else:
        epsilon = 0.58

    sf = 1.25 - 0.25 * (Y_s / 0.184)
    lf = np.log(318) / np.log(Y_abs)
    if method == "fairchild 2010":
        epsilon *= sf * lf
    else:
        epsilon /= sf * lf

    return epsilon


def XYZ_to_hdr_CIELab(
    XYZ: ArrayLike,
    illuminant: ArrayLike = CCS_ILLUMINANTS[
        "CIE 1931 2 Degree Standard Observer"
    ]["D65"],
    Y_s: FloatingOrArrayLike = 0.2,
    Y_abs: FloatingOrArrayLike = 100,
    method: Union[
        Literal["Fairchild 2011", "Fairchild 2010"], str
    ] = "Fairchild 2011",
) -> NDArray:
    """
    Convert from *CIE XYZ* tristimulus values to *hdr-CIELAB* colourspace.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values.
    illuminant
        Reference *illuminant* *CIE xy* chromaticity coordinates or *CIE xyY*
        colourspace array.
    Y_s
        Relative luminance :math:`Y_s` of the surround.
    Y_abs
        Absolute luminance :math:`Y_{abs}` of the scene diffuse white in
        :math:`cd/m^2`.
    method
        Computation method.

    Returns
    -------
    :class:`numpy.ndarray`
        *hdr-CIELAB* colourspace array.

    Notes
    -----
    +----------------+-------------------------+---------------------+
    | **Domain**     | **Scale - Reference**   | **Scale - 1**       |
    +================+=========================+=====================+
    | ``XYZ``        | [0, 1]                  | [0, 1]              |
    +----------------+-------------------------+---------------------+
    | ``illuminant`` | [0, 1]                  | [0, 1]              |
    +----------------+-------------------------+---------------------+
    | ``Y_s``        | [0, 1]                  | [0, 1]              |
    +----------------+-------------------------+---------------------+

    +----------------+-------------------------+---------------------+
    | **Range**      | **Scale - Reference**   | **Scale - 1**       |
    +================+=========================+=====================+
    | ``Lab_hdr``    | ``L_hdr`` : [0, 100]    | ``L_hdr`` : [0, 1]  |
    |                |                         |                     |
    |                | ``a_hdr`` : [-100, 100] | ``a_hdr`` : [-1, 1] |
    |                |                         |                     |
    |                | ``b_hdr`` : [-100, 100] | ``b_hdr`` : [-1, 1] |
    +----------------+-------------------------+---------------------+

    -   Conversion to polar coordinates to compute the *chroma* :math:`C_{hdr}`
        and *hue* :math:`h_{hdr}` correlates can be safely performed with
        :func:`colour.Lab_to_LCHab` definition.
    -   Conversion to cartesian coordinates from the *Lightness*
        :math:`L_{hdr}`, *chroma* :math:`C_{hdr}` and *hue* :math:`h_{hdr}`
        correlates can be safely performed with :func:`colour.LCHab_to_Lab`
        definition.

    References
    ----------
    :cite:`Fairchild2010`, :cite:`Fairchild2011`

    Examples
    --------
    >>> XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
    >>> XYZ_to_hdr_CIELab(XYZ)  # doctest: +ELLIPSIS
    array([ 51.8700206...,  60.4763385...,  32.1455191...])
    >>> XYZ_to_hdr_CIELab(XYZ, method='Fairchild 2010')  # doctest: +ELLIPSIS
    array([  31.9962111...,  128.0076303...,   48.7695230...])
    """

    X, Y, Z = tsplit(to_domain_1(XYZ))
    method = validate_method(method, HDR_CIELAB_METHODS)

    X_n, Y_n, Z_n = tsplit(xyY_to_XYZ(xy_to_xyY(illuminant)))

    if method == "fairchild 2010":
        lightness_callable = lightness_Fairchild2010
    else:
        lightness_callable = lightness_Fairchild2011

    e = exponent_hdr_CIELab(Y_s, Y_abs, method)

    # Domain and range scaling has already be handled.
    with domain_range_scale("ignore"):
        L_hdr = lightness_callable(Y / Y_n, e)
        a_hdr = 5 * (lightness_callable(X / X_n, e) - L_hdr)
        b_hdr = 2 * (L_hdr - lightness_callable(Z / Z_n, e))

    Lab_hdr = tstack([L_hdr, a_hdr, b_hdr])

    return from_range_100(Lab_hdr)


def hdr_CIELab_to_XYZ(
    Lab_hdr: ArrayLike,
    illuminant: ArrayLike = CCS_ILLUMINANTS[
        "CIE 1931 2 Degree Standard Observer"
    ]["D65"],
    Y_s: FloatingOrArrayLike = 0.2,
    Y_abs: FloatingOrArrayLike = 100,
    method: Union[
        Literal["Fairchild 2011", "Fairchild 2010"], str
    ] = "Fairchild 2011",
) -> NDArray:
    """
    Convert from *hdr-CIELAB* colourspace to *CIE XYZ* tristimulus values.

    Parameters
    ----------
    Lab_hdr
        *hdr-CIELAB* colourspace array.
    illuminant
        Reference *illuminant* *CIE xy* chromaticity coordinates or *CIE xyY*
        colourspace array.
    Y_s
        Relative luminance :math:`Y_s` of the surround.
    Y_abs
        Absolute luminance :math:`Y_{abs}` of the scene diffuse white in
        :math:`cd/m^2`.
    method
        Computation method.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ* tristimulus values.

    Notes
    -----
    +----------------+-------------------------+---------------------+
    | **Domain**     | **Scale - Reference**   | **Scale - 1**       |
    +================+=========================+=====================+
    | ``Lab_hdr``    | ``L_hdr`` : [0, 100]    | ``L_hdr`` : [0, 1]  |
    |                |                         |                     |
    |                | ``a_hdr`` : [-100, 100] | ``a_hdr`` : [-1, 1] |
    |                |                         |                     |
    |                | ``b_hdr`` : [-100, 100] | ``b_hdr`` : [-1, 1] |
    +----------------+-------------------------+---------------------+
    | ``illuminant`` | [0, 1]                  | [0, 1]              |
    +----------------+-------------------------+---------------------+
    | ``Y_s``        | [0, 1]                  | [0, 1]              |
    +----------------+-------------------------+---------------------+

    +----------------+-------------------------+---------------------+
    | **Range**      | **Scale - Reference**   | **Scale - 1**       |
    +================+=========================+=====================+
    | ``XYZ``        | [0, 1]                  | [0, 1]              |
    +----------------+-------------------------+---------------------+

    References
    ----------
    :cite:`Fairchild2010`, :cite:`Fairchild2011`

    Examples
    --------
    >>> Lab_hdr = np.array([51.87002062, 60.4763385, 32.14551912])
    >>> hdr_CIELab_to_XYZ(Lab_hdr)  # doctest: +ELLIPSIS
    array([ 0.2065400...,  0.1219722...,  0.0513695...])
    >>> Lab_hdr = np.array([31.99621114, 128.00763036, 48.76952309])
    >>> hdr_CIELab_to_XYZ(Lab_hdr, method='Fairchild 2010')
    ... # doctest: +ELLIPSIS
    array([ 0.2065400...,  0.1219722...,  0.0513695...])
    """

    L_hdr, a_hdr, b_hdr = tsplit(to_domain_100(Lab_hdr))
    method = validate_method(method, HDR_CIELAB_METHODS)

    X_n, Y_n, Z_n = tsplit(xyY_to_XYZ(xy_to_xyY(illuminant)))

    if method == "fairchild 2010":
        luminance_callable = luminance_Fairchild2010
    else:
        luminance_callable = luminance_Fairchild2011

    e = exponent_hdr_CIELab(Y_s, Y_abs, method)

    # Domain and range scaling has already been handled.
    with domain_range_scale("ignore"):
        Y = luminance_callable(L_hdr, e) * Y_n
        X = luminance_callable((a_hdr + 5 * L_hdr) / 5, e) * X_n
        Z = luminance_callable((-b_hdr + 2 * L_hdr) / 2, e) * Z_n

    XYZ = tstack([X, Y, Z])

    return from_range_1(XYZ)



"""
Barten (1999) Contrast Sensitivity Function
===========================================

Defines the *Barten (1999)* contrast sensitivity function:

-   :func:`colour.contrast.contrast_sensitivity_function_Barten1999`

References
----------
-   :cite:`Barten1999` : Barten, P. G. (1999). Contrast Sensitivity of the
    Human Eye and Its Effects on Image Quality. SPIE. doi:10.1117/3.353254
-   :cite:`Barten2003` : Barten, P. G. J. (2003). Formula for the contrast
    sensitivity of the human eye. In Y. Miyake & D. R. Rasmussen (Eds.),
    Proceedings of SPIE (Vol. 5294, pp. 231-238). doi:10.1117/12.537476
-   :cite:`Cowan2004` : Cowan, M., Kennel, G., Maier, T., & Walker, B. (2004).
    Contrast Sensitivity Experiment to Determine the Bit Depth for Digital
    Cinema. SMPTE Motion Imaging Journal, 113(9), 281-292. doi:10.5594/j11549
-   :cite:`InternationalTelecommunicationUnion2015` : International
    Telecommunication Union. (2015). Report ITU-R BT.2246-4 - The present
    state of ultra-high definition television BT Series Broadcasting service
    (Vol. 5, pp. 1-92).
    https://www.itu.int/dms_pub/itu-r/opb/rep/R-REP-BT.2246-4-2015-PDF-E.pdf
"""

from __future__ import annotations

import numpy as np

from colour.hints import (
    Boolean,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    Optional,
)
from colour.utilities import as_float_array, as_float

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "optical_MTF_Barten1999",
    "pupil_diameter_Barten1999",
    "sigma_Barten1999",
    "retinal_illuminance_Barten1999",
    "maximum_angular_size_Barten1999",
    "contrast_sensitivity_function_Barten1999",
]


def optical_MTF_Barten1999(
    u: FloatingOrArrayLike, sigma: FloatingOrArrayLike = 0.01
) -> FloatingOrNDArray:
    """
    Return the optical modulation transfer function (MTF) :math:`M_{opt}` of
    the eye using *Barten (1999)* method.

    Parameters
    ----------
    u
        Spatial frequency :math:`u`, the cycles per degree.
    sigma
        Standard deviation :math:`\\sigma` of the line-spread function
        resulting from the convolution of the different elements of the
        convolution process.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Optical modulation transfer function (MTF) :math:`M_{opt}` of the eye.

    References
    ----------
    :cite:`Barten1999`, :cite:`Barten2003`, :cite:`Cowan2004`,
    :cite:`InternationalTelecommunicationUnion2015`,

    Examples
    --------
    >>> optical_MTF_Barten1999(4, 0.01)  # doctest: +ELLIPSIS
    0.9689107...
    """

    u = as_float_array(u)
    sigma = as_float_array(sigma)

    return as_float(np.exp(-2 * np.pi**2 * sigma**2 * u**2))


def pupil_diameter_Barten1999(
    L: FloatingOrArrayLike,
    X_0: FloatingOrArrayLike = 60,
    Y_0: Optional[FloatingOrArrayLike] = None,
) -> FloatingOrNDArray:
    """
    Return the pupil diameter for given luminance and object or stimulus
    angular size using *Barten (1999)* method.

    Parameters
    ----------
    L
        Average luminance :math:`L` in :math:`cd/m^2`.
    X_0
        Angular size of the object :math:`X_0` in degrees in the x direction.
    Y_0
        Angular size of the object :math:`X_0` in degrees in the y direction.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Pupil diameter.

    References
    ----------
    :cite:`Barten1999`, :cite:`Barten2003`, :cite:`Cowan2004`,
    :cite:`InternationalTelecommunicationUnion2015`,

    Examples
    --------
    >>> pupil_diameter_Barten1999(100, 60, 60)  # doctest: +ELLIPSIS
    2.0777571...
    """

    L = as_float_array(L)
    X_0 = as_float_array(X_0)
    Y_0 = X_0 if Y_0 is None else as_float_array(Y_0)

    return as_float(5 - 3 * np.tanh(0.4 * np.log(L * X_0 * Y_0 / 40**2)))


def sigma_Barten1999(
    sigma_0: FloatingOrArrayLike = 0.5 / 60,
    C_ab: FloatingOrArrayLike = 0.08 / 60,
    d: FloatingOrArrayLike = 2.1,
) -> FloatingOrNDArray:
    """
    Return the standard deviation :math:`\\sigma` of the line-spread function
    resulting from the convolution of the different elements of the convolution
    process using *Barten (1999)* method.

    The :math:`\\sigma` quantity depends on the pupil diameter :math:`d` of the
    eye lens. For very small pupil diameters, :math:`\\sigma` increases
    inversely proportionally with pupil size because of diffraction, and for
    large pupil diameters, :math:`\\sigma` increases about linearly with pupil
    size because of chromatic aberration and others aberrations.

    Parameters
    ----------
    sigma_0
        Constant :math:`\\sigma_{0}` in degrees.
    C_ab
        Spherical aberration of the eye :math:`C_{ab}` in
        :math:`degrees\\div mm`.
    d
        Pupil diameter :math:`d` in millimeters.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Standard deviation :math:`\\sigma` of the line-spread function
        resulting from the convolution of the different elements of the
        convolution process.

    Warnings
    --------
    This definition expects :math:`\\sigma_{0}` and :math:`C_{ab}` to be given
    in degrees and :math:`degrees\\div mm` respectively. However, in the
    literature, the values for :math:`\\sigma_{0}` and
    :math:`C_{ab}` are usually given in :math:`arc min` and
    :math:`arc min\\div mm` respectively, thus they need to be divided by 60.

    References
    ----------
    :cite:`Barten1999`, :cite:`Barten2003`, :cite:`Cowan2004`,
    :cite:`InternationalTelecommunicationUnion2015`,

    Examples
    --------
    >>> sigma_Barten1999(0.5 / 60, 0.08 / 60, 2.1)  # doctest: +ELLIPSIS
    0.0087911...
    """

    sigma_0 = as_float_array(sigma_0)
    C_ab = as_float_array(C_ab)
    d = as_float_array(d)

    return as_float(np.sqrt(sigma_0**2 + (C_ab * d) ** 2))


def retinal_illuminance_Barten1999(
    L: FloatingOrArrayLike,
    d: FloatingOrArrayLike = 2.1,
    apply_stiles_crawford_effect_correction: Boolean = True,
) -> FloatingOrNDArray:
    """
    Return the retinal illuminance :math:`E` in Trolands for given average
    luminance :math:`L` and pupil diameter :math:`d` using *Barten (1999)*
    method.

    Parameters
    ----------
    L
        Average luminance :math:`L` in :math:`cd/m^2`.
    d
        Pupil diameter :math:`d` in millimeters.
    apply_stiles_crawford_effect_correction
        Whether to apply the correction for *Stiles-Crawford* effect.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Retinal illuminance :math:`E` in Trolands.

    Notes
    -----
    -   This definition is for use with photopic viewing conditions and thus
        corrects for the Stiles-Crawford effect by default, i.e. directional
        sensitivity of the cone cells with lower response of cone cells
        receiving light from the edge of the pupil.

    References
    ----------
    :cite:`Barten1999`, :cite:`Barten2003`, :cite:`Cowan2004`,
    :cite:`InternationalTelecommunicationUnion2015`,

    Examples
    --------
    >>> retinal_illuminance_Barten1999(100, 2.1)  # doctest: +ELLIPSIS
    330.4115803...
    >>> retinal_illuminance_Barten1999(100, 2.1, False)  # doctest: +ELLIPSIS
    346.3605900...
    """

    d = as_float_array(d)
    L = as_float_array(L)

    E = (np.pi * d**2) / 4 * L

    if apply_stiles_crawford_effect_correction:
        E *= 1 - (d / 9.7) ** 2 + (d / 12.4) ** 4

    return as_float(E)


def maximum_angular_size_Barten1999(
    u: FloatingOrArrayLike,
    X_0: FloatingOrArrayLike = 60,
    X_max: FloatingOrArrayLike = 12,
    N_max: FloatingOrArrayLike = 15,
) -> FloatingOrNDArray:
    """
    Return the maximum angular size :math:`X` of the object considered using
    *Barten (1999)* method.

    Parameters
    ----------
    u
        Spatial frequency :math:`u`, the cycles per degree.
    X_0
        Angular size :math:`X_0` in degrees of the object in the x direction.
    X_max
        Maximum angular size :math:`X_{max}` in degrees of the integration
        area in the x direction.
    N_max
        Maximum number of cycles :math:`N_{max}` over which the eye can
        integrate the information.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Maximum angular size :math:`X` of the object considered.

    References
    ----------
    :cite:`Barten1999`, :cite:`Barten2003`, :cite:`Cowan2004`,
    :cite:`InternationalTelecommunicationUnion2015`,

    Examples
    --------
    >>> maximum_angular_size_Barten1999(4)  # doctest: +ELLIPSIS
    3.5729480...
    """

    u = as_float_array(u)
    X_0 = as_float_array(X_0)
    X_max = as_float_array(X_max)
    N_max = as_float_array(N_max)

    return as_float(
        (1 / X_0**2 + 1 / X_max**2 + u**2 / N_max**2) ** -0.5
    )


def contrast_sensitivity_function_Barten1999(
    u: FloatingOrArrayLike,
    sigma: FloatingOrArrayLike = sigma_Barten1999(0.5 / 60, 0.08 / 60, 2.1),
    k: FloatingOrArrayLike = 3.0,
    T: FloatingOrArrayLike = 0.1,
    X_0: FloatingOrArrayLike = 60,
    Y_0: Optional[FloatingOrArrayLike] = None,
    X_max: FloatingOrArrayLike = 12,
    Y_max: Optional[FloatingOrArrayLike] = None,
    N_max: FloatingOrArrayLike = 15,
    n: FloatingOrArrayLike = 0.03,
    p: FloatingOrArrayLike = 1.2274 * 10**6,
    E: FloatingOrArrayLike = retinal_illuminance_Barten1999(20, 2.1),
    phi_0: FloatingOrArrayLike = 3 * 10**-8,
    u_0: FloatingOrArrayLike = 7,
) -> FloatingOrNDArray:
    """
    Return the contrast sensitivity :math:`S` of the human eye according to
    the contrast sensitivity function (CSF) described by *Barten (1999)*.

    Contrast sensitivity is defined as the inverse of the modulation threshold
    of a sinusoidal luminance pattern. The modulation threshold of this pattern
    is generally defined by 50% probability of detection. The contrast
    sensitivity function or CSF gives the contrast sensitivity as a function of
    spatial frequency. In the CSF, the spatial frequency is expressed in
    angular units with respect to the eye. It reaches a maximum between 1 and
    10 cycles per degree with a fall off at higher and lower spatial
    frequencies.

    Parameters
    ----------
    u
        Spatial frequency :math:`u`, the cycles per degree.
    sigma
        Standard deviation :math:`\\sigma` of the line-spread function
        resulting from the convolution of the different elements of the
        convolution process.
    k
        Signal-to-noise (SNR) ratio :math:`k`.
    T
        Integration time :math:`T` in seconds of the eye.
    X_0
        Angular size :math:`X_0` in degrees of the object in the x direction.
    Y_0
        Angular size :math:`Y_0` in degrees of the object in the y direction.
    X_max
        Maximum angular size :math:`X_{max}` in degrees of the integration
        area in the x direction.
    Y_max
        Maximum angular size :math:`Y_{max}` in degrees of the integration
        area in the y direction.
    N_max
        Maximum number of cycles :math:`N_{max}` over which the eye can
        integrate the information.
    n
        Quantum efficiency of the eye :math:`n`.
    p
        Photon conversion factor :math:`p` in
        :math:`photons\\div seconds\\div degrees^2\\div Trolands` that
        depends on the light source.
    E
        Retinal illuminance :math:`E` in Trolands.
    phi_0
        Spectral density :math:`\\phi_0` in :math:`seconds degrees^2` of the
        neural noise.
    u_0
        Spatial frequency :math:`u_0` in :math:`cycles\\div degrees` above
        which the lateral inhibition ceases.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Contrast sensitivity :math:`S`.

    Warnings
    --------
    This definition expects :math:`\\sigma_{0}` and :math:`C_{ab}` used in the
    computation of :math:`\\sigma` to be given in degrees and
    :math:`degrees\\div mm` respectively. However, in the literature, the
    values for :math:`\\sigma_{0}` and :math:`C_{ab}` are usually given in
    :math:`arc min` and :math:`arc min\\div mm` respectively, thus they need to
    be divided by 60.

    Notes
    -----
    -   The formula holds for bilateral viewing and for equal dimensions of
        the object in x and y direction. For monocular vision, the contrast
        sensitivity is a factor :math:`\\sqrt{2}` smaller.
    -   *Barten (1999)* CSF default values for the :math:`k`,
        :math:`\\sigma_{0}`, :math:`C_{ab}`, :math:`T`, :math:`X_{max}`,
        :math:`N_{max}`, :math:`n`, :math:`\\phi_{0}` and :math:`u_0` constants
        are valid for a standard observer with good vision and with an age
        between 20 and 30 years.
    -   The other constants have been filled using reference data from
        *Figure 31* in :cite:`InternationalTelecommunicationUnion2015` but
        must be adapted to the current use case.
    -   The product of :math:`u`, the cycles per degree, and :math:`X_0`,
        the number of degrees, gives the number of cycles :math:`P_c` in a
        pattern. Therefore, :math:`X_0` can be made a variable dependent on
        :math:`u` such as :math:`X_0 = P_c / u`.

    References
    ----------
    :cite:`Barten1999`, :cite:`Barten2003`, :cite:`Cowan2004`,
    :cite:`InternationalTelecommunicationUnion2015`,

    Examples
    --------
    >>> contrast_sensitivity_function_Barten1999(4)  # doctest: +ELLIPSIS
    360.8691122...

    Reproducing *Figure 31* in \
:cite:`InternationalTelecommunicationUnion2015` illustrating the minimum
    detectable contrast according to *Barten (1999)* model with the assumed
    conditions for UHDTV applications. The minimum detectable contrast
    :math:`MDC` is then defined as follows::

        :math:`MDC = 1 / CSF * 2 * (1 / 1.27)`

    where :math:`2` is used for the conversion from modulation to contrast and
    :math:`1 / 1.27` is used for the conversion from sinusoidal to rectangular
    waves.

    >>> from scipy.optimize import fmin
    >>> settings_BT2246 = {
    ...     'k': 3.0,
    ...     'T': 0.1,
    ...     'X_max': 12,
    ...     'N_max': 15,
    ...     'n': 0.03,
    ...     'p': 1.2274 * 10 ** 6,
    ...     'phi_0': 3 * 10 ** -8,
    ...     'u_0': 7,
    ... }
    >>>
    >>> def maximise_spatial_frequency(L):
    ...     maximised_spatial_frequency = []
    ...     for L_v in L:
    ...         X_0 = 60
    ...         d = pupil_diameter_Barten1999(L_v, X_0)
    ...         sigma = sigma_Barten1999(0.5 / 60, 0.08 / 60, d)
    ...         E = retinal_illuminance_Barten1999(L_v, d, True)
    ...         maximised_spatial_frequency.append(
    ...             fmin(lambda x: (
    ...                     -contrast_sensitivity_function_Barten1999(
    ...                         u=x,
    ...                         sigma=sigma,
    ...                         X_0=X_0,
    ...                         E=E,
    ...                         **settings_BT2246)
    ...                 ), 0, disp=False)[0])
    ...     return as_float(np.array(maximised_spatial_frequency))
    >>>
    >>> L = np.logspace(np.log10(0.01), np.log10(100), 10)
    >>> X_0 = Y_0 = 60
    >>> d = pupil_diameter_Barten1999(L, X_0, Y_0)
    >>> sigma = sigma_Barten1999(0.5 / 60, 0.08 / 60, d)
    >>> E = retinal_illuminance_Barten1999(L, d)
    >>> u = maximise_spatial_frequency(L)
    >>> (1 / contrast_sensitivity_function_Barten1999(
    ...     u=u, sigma=sigma, E=E, X_0=X_0, Y_0=Y_0, **settings_BT2246)
    ...  * 2 * (1/ 1.27))
    ... # doctest: +ELLIPSIS
    array([ 0.0207396...,  0.0133019...,  0.0089256...,  0.0064202...,  \
0.0050275...,
            0.0041933...,  0.0035573...,  0.0030095...,  0.0025803...,  \
0.0022897...])
    """

    u = as_float_array(u)
    k = as_float_array(k)
    T = as_float_array(T)
    X_0 = as_float_array(X_0)
    Y_0 = X_0 if Y_0 is None else as_float_array(Y_0)
    X_max = as_float_array(X_max)
    Y_max = X_max if Y_max is None else as_float_array(Y_max)
    N_max = as_float_array(N_max)
    n = as_float_array(n)
    p = as_float_array(p)
    E = as_float_array(E)
    phi_0 = as_float_array(phi_0)
    u_0 = as_float_array(u_0)

    M_opt = optical_MTF_Barten1999(u, sigma)

    M_as = 1 / (
        maximum_angular_size_Barten1999(u, X_0, X_max, N_max)
        * maximum_angular_size_Barten1999(u, Y_0, Y_max, N_max)
    )

    S = (M_opt / k) / np.sqrt(
        2
        / T
        * M_as
        * (1 / (n * p * E) + phi_0 / (1 - np.exp(-((u / u_0) ** 2))))
    )

    return as_float(S)



"""Showcases *Hunt* colour appearance model computations."""

import numpy as np

import colour
from colour.appearance.hunt import CAM_ReferenceSpecification_Hunt
from colour.utilities import message_box

message_box('"Hunt" Colour Appearance Model Computations')

XYZ = np.array([19.01, 20.00, 21.78])
XYZ_w = np.array([95.05, 100.00, 108.88])
XYZ_b = np.array([95.05, 100.00, 108.88])
L_A = 318.31
surround = colour.VIEWING_CONDITIONS_HUNT["Normal Scenes"]
CCT_w = 6504.0
message_box(
    f'Converting to the "Hunt" colour appearance model specification using '
    f"given parameters:\n\n"
    f"\tXYZ: {XYZ}\n"
    f"\tXYZ_w: {XYZ_w}\n"
    f"\tXYZ_b: {XYZ_b}\n"
    f"\tL_A: {L_A}\n"
    f"\tsurround: {surround}\n"
    f"\tCCT_w: {CCT_w}"
)

specification = colour.XYZ_to_Hunt(
    XYZ, XYZ_w, XYZ_b, L_A, surround, CCT_w=CCT_w
)
print(specification)

print("\n")

message_box(
    'Broadcasting the current output "Hunt" colour appearance '
    "model specification to the reference specification.\n"
    "The intent of this reference specification is to provide names "
    'as closest as possible to the "Mark D. Fairchild" reference.\n'
    "The current output specification is meant to be consistent with "
    "the other colour appearance model specification by using same "
    "argument names for consistency wherever possible."
)

print(CAM_ReferenceSpecification_Hunt(*specification.values))



"""
Corresponding Chromaticities Prediction Plotting
================================================

Defines the corresponding chromaticities prediction plotting objects:

-   :func:`colour.plotting.plot_corresponding_chromaticities_prediction`
"""

from __future__ import annotations

import matplotlib.pyplot as plt

from colour.corresponding import (
    CorrespondingColourDataset,
    corresponding_chromaticities_prediction,
)
from colour.hints import Any, Dict, Literal, Optional, Tuple, Union
from colour.plotting import (
    CONSTANTS_COLOUR_STYLE,
    artist,
    plot_chromaticity_diagram_CIE1976UCS,
    override_style,
    render,
)
from colour.utilities import is_numeric

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "plot_corresponding_chromaticities_prediction",
]


@override_style()
def plot_corresponding_chromaticities_prediction(
    experiment: Union[
        Literal[1, 2, 3, 4, 6, 8, 9, 11, 12], CorrespondingColourDataset
    ] = 1,
    model: Union[
        Literal[
            "CIE 1994",
            "CMCCAT2000",
            "Fairchild 1990",
            "Von Kries",
            "Zhai 2018",
        ],
        str,
    ] = "Von Kries",
    corresponding_chromaticities_prediction_kwargs: Optional[Dict] = None,
    **kwargs: Any,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Plot given chromatic adaptation model corresponding chromaticities
    prediction.

    Parameters
    ----------
    experiment
        *Breneman (1987)* experiment number or
        :class:`colour.CorrespondingColourDataset` class instance.
    model
        Corresponding chromaticities prediction model name.
    corresponding_chromaticities_prediction_kwargs
        Keyword arguments for the :func:`colour.\
corresponding_chromaticities_prediction` definition.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.plotting.artist`,
        :func:`colour.plotting.diagrams.plot_chromaticity_diagram`,
        :func:`colour.plotting.render`},
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.

    Examples
    --------
    >>> plot_corresponding_chromaticities_prediction(1, 'Von Kries')
    ... # doctest: +ELLIPSIS
    (<Figure size ... with 1 Axes>, <...AxesSubplot...>)

    .. image:: ../_static/Plotting_\
Plot_Corresponding_Chromaticities_Prediction.png
        :align: center
        :alt: plot_corresponding_chromaticities_prediction
    """

    if corresponding_chromaticities_prediction_kwargs is None:
        corresponding_chromaticities_prediction_kwargs = {}

    settings: Dict[str, Any] = {"uniform": True}
    settings.update(kwargs)

    _figure, axes = artist(**settings)

    name = (
        f"Experiment {experiment}"
        if is_numeric(experiment)
        else experiment.name  # type: ignore[union-attr]
    )
    title = (
        f"Corresponding Chromaticities Prediction - {model} - {name} - "
        "CIE 1976 UCS Chromaticity Diagram"
    )

    settings = {"axes": axes, "title": title}
    settings.update(kwargs)
    settings["standalone"] = False

    plot_chromaticity_diagram_CIE1976UCS(**settings)

    results = corresponding_chromaticities_prediction(
        experiment, model, **corresponding_chromaticities_prediction_kwargs
    )

    for result in results:
        _name, uv_t, uv_m, uv_p = result
        axes.arrow(
            uv_t[0],
            uv_t[1],
            uv_p[0] - uv_t[0] - 0.1 * (uv_p[0] - uv_t[0]),
            uv_p[1] - uv_t[1] - 0.1 * (uv_p[1] - uv_t[1]),
            color=CONSTANTS_COLOUR_STYLE.colour.dark,
            head_width=0.005,
            head_length=0.005,
            zorder=CONSTANTS_COLOUR_STYLE.zorder.midground_annotation,
        )
        axes.plot(
            uv_t[0],
            uv_t[1],
            "o",
            color=CONSTANTS_COLOUR_STYLE.colour.brightest,
            markeredgecolor=CONSTANTS_COLOUR_STYLE.colour.dark,
            markersize=(
                CONSTANTS_COLOUR_STYLE.geometry.short * 6
                + CONSTANTS_COLOUR_STYLE.geometry.short * 0.75
            ),
            markeredgewidth=CONSTANTS_COLOUR_STYLE.geometry.short * 0.75,
            zorder=CONSTANTS_COLOUR_STYLE.zorder.foreground_line,
        )
        axes.plot(
            uv_m[0],
            uv_m[1],
            "^",
            color=CONSTANTS_COLOUR_STYLE.colour.brightest,
            markeredgecolor=CONSTANTS_COLOUR_STYLE.colour.dark,
            markersize=(
                CONSTANTS_COLOUR_STYLE.geometry.short * 6
                + CONSTANTS_COLOUR_STYLE.geometry.short * 0.75
            ),
            markeredgewidth=CONSTANTS_COLOUR_STYLE.geometry.short * 0.75,
            zorder=CONSTANTS_COLOUR_STYLE.zorder.foreground_line,
        )
        axes.plot(
            uv_p[0],
            uv_p[1],
            "^",
            color=CONSTANTS_COLOUR_STYLE.colour.dark,
            zorder=CONSTANTS_COLOUR_STYLE.zorder.foreground_line,
        )

    settings.update(
        {
            "standalone": True,
            "bounding_box": (-0.1, 0.7, -0.1, 0.7),
        }
    )
    settings.update(kwargs)

    return render(**settings)



"""
:math:`I_GP_GT_G` Colourspace
=============================

Defines the :math:`I_GP_GT_G` colourspace transformations:

-   :func:`colour.XYZ_to_IgPgTg`
-   :func:`colour.IgPgTg_to_XYZ`

References
----------
-   :cite:`Hellwig2020` : Hellwig, L., & Fairchild, M. D. (2020). Using
    Gaussian Spectra to Derive a Hue-linear Color Space. Journal of Perceptual
    Imaging. doi:10.2352/J.Percept.Imaging.2020.3.2.020401
"""

from __future__ import annotations

import numpy as np

from colour.algebra import spow, vector_dot
from colour.hints import ArrayLike, NDArray
from colour.utilities import from_range_1, to_domain_1

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "MATRIX_IGPGTG_XYZ_TO_LMS",
    "MATRIX_IGPGTG_LMS_TO_XYZ",
    "MATRIX_IGPGTG_LMS_P_TO_IGPGTG",
    "MATRIX_IGPGTG_IGPGTG_TO_LMS_P",
    "XYZ_to_IgPgTg",
    "IgPgTg_to_XYZ",
]

MATRIX_IGPGTG_XYZ_TO_LMS: NDArray = np.array(
    [
        [2.968, 2.741, -0.649],
        [1.237, 5.969, -0.173],
        [-0.318, 0.387, 2.311],
    ]
)
"""*CIE XYZ* tristimulus values to normalised cone responses matrix."""

MATRIX_IGPGTG_LMS_TO_XYZ: NDArray = np.linalg.inv(MATRIX_IGPGTG_XYZ_TO_LMS)
"""Normalised cone responses to *CIE XYZ* tristimulus values matrix."""

MATRIX_IGPGTG_LMS_P_TO_IGPGTG: NDArray = np.array(
    [
        [0.117, 1.464, 0.130],
        [8.285, -8.361, 21.400],
        [-1.208, 2.412, -36.530],
    ]
)
"""Normalised non-linear cone responses to :math:`I_GP_GT_G` colourspace matrix."""

MATRIX_IGPGTG_IGPGTG_TO_LMS_P: NDArray = np.linalg.inv(
    MATRIX_IGPGTG_LMS_P_TO_IGPGTG
)
""":math:`I_GP_GT_G` colourspace to normalised non-linear cone responses matrix."""


def XYZ_to_IgPgTg(XYZ: ArrayLike) -> NDArray:
    """
    Convert from *CIE XYZ* tristimulus values to :math:`I_GP_GT_G`
    colourspace.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values.

    Returns
    -------
    :class:`numpy.ndarray`
        :math:`I_GP_GT_G` colourspace array.

    Notes
    -----
    +------------+-----------------------+-----------------+
    | **Domain** | **Scale - Reference** | **Scale - 1**   |
    +============+=======================+=================+
    | ``XYZ``    | [0, 1]                | [0, 1]          |
    +------------+-----------------------+-----------------+

    +------------+-----------------------+-----------------+
    | **Range**  | **Scale - Reference** | **Scale - 1**   |
    +============+=======================+=================+
    | ``IgPgTg`` | ``IG`` : [0, 1]       | ``IG`` : [0, 1] |
    |            |                       |                 |
    |            | ``PG`` : [-1, 1]      | ``PG`` : [-1, 1]|
    |            |                       |                 |
    |            | ``TG`` : [-1, 1]      | ``TG`` : [-1, 1]|
    +------------+-----------------------+-----------------+

    -   Input *CIE XYZ* tristimulus values must be adapted to
        *CIE Standard Illuminant D Series* *D65*.

    References
    ----------
    :cite:`Hellwig2020`

    Examples
    --------
    >>> XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
    >>> XYZ_to_IgPgTg(XYZ)  # doctest: +ELLIPSIS
    array([ 0.4242125...,  0.1863249...,  0.1068922...])
    """

    XYZ = to_domain_1(XYZ)

    LMS = vector_dot(MATRIX_IGPGTG_XYZ_TO_LMS, XYZ)
    LMS_prime = spow(LMS / np.array([18.36, 21.46, 19435]), 0.427)
    IgPgTg = vector_dot(MATRIX_IGPGTG_LMS_P_TO_IGPGTG, LMS_prime)

    return from_range_1(IgPgTg)


def IgPgTg_to_XYZ(IgPgTg: ArrayLike) -> NDArray:
    """
    Convert from :math:`I_GP_GT_G` colourspace to *CIE XYZ* tristimulus
    values.

    Parameters
    ----------
    IgPgTg
        :math:`I_GP_GT_G` colourspace array.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ* tristimulus values.

    Notes
    -----
    +------------+-----------------------+-----------------+
    | **Domain** | **Scale - Reference** | **Scale - 1**   |
    +============+=======================+=================+
    | ``IgPgTg`` | ``IG`` : [0, 1]       | ``IG`` : [0, 1] |
    |            |                       |                 |
    |            | ``PG`` : [-1, 1]      | ``PG`` : [-1, 1]|
    |            |                       |                 |
    |            | ``TG`` : [-1, 1]      | ``TG`` : [-1, 1]|
    +------------+-----------------------+-----------------+

    +------------+-----------------------+-----------------+
    | **Range**  | **Scale - Reference** | **Scale - 1**   |
    +============+=======================+=================+
    | ``XYZ``    | [0, 1]                | [0, 1]          |
    +------------+-----------------------+-----------------+

    References
    ----------
    :cite:`Hellwig2020`

    Examples
    --------
    >>> IgPgTg = np.array([0.42421258, 0.18632491, 0.10689223])
    >>> IgPgTg_to_XYZ(IgPgTg)  # doctest: +ELLIPSIS
    array([ 0.2065400...,  0.1219722...,  0.0513695...])
    """

    IgPgTg = to_domain_1(IgPgTg)

    LMS = vector_dot(MATRIX_IGPGTG_IGPGTG_TO_LMS_P, IgPgTg)
    LMS_prime = spow(LMS, 1 / 0.427) * np.array([18.36, 21.46, 19435])
    XYZ = vector_dot(MATRIX_IGPGTG_LMS_TO_XYZ, LMS_prime)

    return from_range_1(XYZ)



"""Showcases common plotting examples."""

from colour.plotting import (
    ColourSwatch,
    colour_style,
    plot_multi_colour_swatches,
    plot_single_colour_swatch,
)
from colour.utilities import message_box

message_box("Common Plots")

colour_style()

message_box("Plotting a single colour.")
plot_single_colour_swatch(
    ColourSwatch((0.32315746, 0.32983556, 0.33640183), "Neutral 5 (.70 D)"),
    text_size=32,
)

print("\n")

message_box("Plotting multiple colours.")
plot_multi_colour_swatches(
    (
        ColourSwatch((0.45293517, 0.31732158, 0.26414773), "Dark Skin"),
        ColourSwatch((0.77875824, 0.57726450, 0.50453169), "Light Skin"),
    ),
    text_size=32,
)



"""
Spectral Bandpass Dependence Correction
=======================================

Defines the objects to perform spectral bandpass dependence correction.

The following correction methods are available:

-   :func:`colour.colorimetry.bandpass_correction_Stearns1988`:
    *Stearns and Stearns (1988)* spectral bandpass dependence correction
    method.
-   :attr:`colour.BANDPASS_CORRECTION_METHODS`: Supported spectral bandpass
    dependence correction methods.
-   :func:`colour.bandpass_correction`: Spectral bandpass dependence
    correction using given method.

References
----------
-   :cite:`Stearns1988a` : Stearns, E. I., & Stearns, R. E. (1988). An example
    of a method for correcting radiance data for Bandpass error. Color Research
    & Application, 13(4), 257-259. doi:10.1002/col.5080130410
-   :cite:`Westland2012f` : Westland, S., Ripamonti, C., & Cheung, V. (2012).
    Correction for Spectral Bandpass. In Computational Colour Science Using
    MATLAB (2nd ed., p. 38). ISBN:978-0-470-66569-5
"""

from __future__ import annotations

import numpy as np

from colour.colorimetry import SpectralDistribution
from colour.hints import Floating, Literal, Union
from colour.utilities import CaseInsensitiveMapping, validate_method

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "bandpass_correction_Stearns1988",
    "BANDPASS_CORRECTION_METHODS",
    "bandpass_correction",
]

CONSTANT_ALPHA_STEARNS: Floating = 0.083


def bandpass_correction_Stearns1988(
    sd: SpectralDistribution,
) -> SpectralDistribution:
    """
    Implement spectral bandpass dependence correction on given spectral
    distribution using *Stearns and Stearns (1988)* method.

    Parameters
    ----------
    sd
        Spectral distribution.

    Returns
    -------
    :class:`colour.SpectralDistribution`
        Spectral bandpass dependence corrected spectral distribution.

    References
    ----------
    :cite:`Stearns1988a`, :cite:`Westland2012f`

    Examples
    --------
    >>> from colour import SpectralDistribution
    >>> from colour.utilities import numpy_print_options
    >>> data = {
    ...     500: 0.0651,
    ...     520: 0.0705,
    ...     540: 0.0772,
    ...     560: 0.0870,
    ...     580: 0.1128,
    ...     600: 0.1360
    ... }
    >>> with numpy_print_options(suppress=True):
    ...     bandpass_correction_Stearns1988(SpectralDistribution(data))
    ... # doctest: +ELLIPSIS
    SpectralDistribution([[ 500.        ,    0.0646518...],
                          [ 520.        ,    0.0704293...],
                          [ 540.        ,    0.0769485...],
                          [ 560.        ,    0.0856928...],
                          [ 580.        ,    0.1129644...],
                          [ 600.        ,    0.1379256...]],
                         interpolator=SpragueInterpolator,
                         interpolator_kwargs={},
                         extrapolator=Extrapolator,
                         extrapolator_kwargs={...})
    """

    values = np.copy(sd.values)
    values[0] = (1 + CONSTANT_ALPHA_STEARNS) * values[
        0
    ] - CONSTANT_ALPHA_STEARNS * values[1]
    values[-1] = (1 + CONSTANT_ALPHA_STEARNS) * values[
        -1
    ] - CONSTANT_ALPHA_STEARNS * values[-2]
    for i in range(1, len(values) - 1):
        values[i] = (
            -CONSTANT_ALPHA_STEARNS * values[i - 1]
            + (1 + 2 * CONSTANT_ALPHA_STEARNS) * values[i]
            - CONSTANT_ALPHA_STEARNS * values[i + 1]
        )

    sd.values = values

    return sd


BANDPASS_CORRECTION_METHODS: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {"Stearns 1988": bandpass_correction_Stearns1988}
)
BANDPASS_CORRECTION_METHODS.__doc__ = """
Supported spectral bandpass dependence correction methods.
"""


def bandpass_correction(
    sd: SpectralDistribution,
    method: Union[Literal["Stearns 1988"], str] = "Stearns 1988",
) -> SpectralDistribution:
    """
    Implement spectral bandpass dependence correction on given spectral
    distribution using given method.

    Parameters
    ----------
    sd
        Spectral distribution.
    method
        Correction method.

    Returns
    -------
    :class:`colour.SpectralDistribution`
        Spectral bandpass dependence corrected spectral distribution.

    References
    ----------
    :cite:`Stearns1988a`, :cite:`Westland2012f`

    Examples
    --------
    >>> from colour import SpectralDistribution
    >>> from colour.utilities import numpy_print_options
    >>> data = {
    ...     500: 0.0651,
    ...     520: 0.0705,
    ...     540: 0.0772,
    ...     560: 0.0870,
    ...     580: 0.1128,
    ...     600: 0.1360
    ... }
    >>> with numpy_print_options(suppress=True):
    ...     bandpass_correction(SpectralDistribution(data))
    ... # doctest: +ELLIPSIS
    SpectralDistribution([[ 500.        ,    0.0646518...],
                          [ 520.        ,    0.0704293...],
                          [ 540.        ,    0.0769485...],
                          [ 560.        ,    0.0856928...],
                          [ 580.        ,    0.1129644...],
                          [ 600.        ,    0.1379256...]],
                         interpolator=SpragueInterpolator,
                         interpolator_kwargs={},
                         extrapolator=Extrapolator,
                         extrapolator_kwargs={...})
    """

    method = validate_method(method, BANDPASS_CORRECTION_METHODS)

    return BANDPASS_CORRECTION_METHODS[method](sd)
